
//*********************************************************
//
//  UNIT:     C o l o r R G B          04.04.95, 18.09.2006
//
//  color definitions, routines and conversion for Delphi-5/6/7
//  coded by Gerd Platl
//---------------------------------------------------------
//  Literature:
//  - Color Conversion Algorithms
//      http://www.cs.rit.edu/~ncs/color/t_convert.html
//---------------------------------------------------------
//  Color32          create a new tColor32
//  RandomColor      create a random color
//  WinColor         convert arguments to tColor value
//  Color2RGB        convert color value -> RGB value
//  ShadeColor       return color with given intensity
//  MixColor         mix 2 colors with given intensity
//  GrayValue        return gray value of color
//
//  LongColor        convert color -> long color
//  LongColor2Color  convert long color -> color
//
//  RGBRealToByte    convert real color -> color
//  RGBByteToReal    convert color -> real color
//  RGBRealGray      return gray value of real color
//  RGBNorm          nomalize real color
//  RGBRandom        return random real color
//  ColorString      convert color values to string
//  ColorShortString convert color values to short string
//  ColorHexString   convert color values to hex string
//
//*********************************************************
{$I comp_opt.inc}

UNIT ColorRGB;

INTERFACE

{$ifdef graph32}
  uses  Graphics, SysUtil2, Gr32;
{$else}
  uses  Graphics, SysUtil2;
{$endif}

{$I float.inc}

{$I Colors32.inc                        color constant list}

const maxColorRGB = 255;               {global color contants}
      alphaMask = $FF000000;
      redMask   = $00FF0000;
      greenMask = $0000FF00;
      blueMask  = $000000FF;
      rgbMask   = $00FFFFFF;

//=== t C o l o r 3 2 ===
                                       {transparency:}
const trOpaque    = 0;                 { = undurchsichtig}
      trInvisible = 255;               { = durchsichtig}

{$ifdef graph32}
  type  tColor32 = Gr32.tColor32;   {A R G B}
        tArrayOfColor32 = Gr32.tArrayOfColor32;    
{$else}
  type  tColor32 = Cardinal;        {A R G B}
        tArrayOfColor32 = array of tColor32;    // entspricht Gr32.TArrayOfColor32
{$endif}

type  pColor32 = ^tColor32;
      tColor32Array = array [0..9999] of tColor32;
      pColor32Array = ^tColor32Array;

type  pColorBGRA = ^tColorBGRA;  // entspricht Gr32.tColor32
      tColorBGRA = packed record   b,g,r,a: byte;   end;

type  pColorRGBI = ^tColorRGBI;  // entspricht Graphics.tColor
      tColorRGBI = packed record   r,g,b,a: byte;   end;

//=== t L o n g C o l o r ===
type  tLongColor = record              {color with 24.8 fix point values}
        r,g,b: longint;   end;         {  rgb range: 0.0 .. 1.0}

//=== t R e a l C o l o r ===
type  pRealColor = ^tRealColor;
      tRealColor = record              {color: red-green-blue}
        r,g,b: float;   end;           {  rgb range: 0.0 .. 1.0}

const maxRealByte = 255.99;
      blackRealColor: tRealColor = (r: 0.0; g: 0.0; b: 0.0);
      redRealColor:   tRealColor = (r: 1.0; g: 0.0; b: 0.0);
      greenRealColor: tRealColor = (r: 0.0; g: 1.0; b: 0.0);
      blueRealColor:  tRealColor = (r: 0.0; g: 0.0; b: 1.0);
      whiteRealColor: tRealColor = (r: 1.0; g: 1.0; b: 1.0);

{ 8 x 8 - dither matrix for 64 intensity patterns }
const intMatrix8: array [0..7,0..7] of byte =
     (( 0,32, 8,40, 2,34,10,42),(48,16,56,24,50,18,58,26)
     ,(12,44, 4,36,14,46, 6,38),(60,28,52,20,62,30,54,22)
     ,( 3,35,11,43, 1,33, 9,41),(51,19,59,27,49,17,57,25)
     ,(15,47, 7,39,13,45, 5,37),(63,31,55,23,61,29,53,21));

//=== t C o l o r ===   {declared in standard unit Graphics}
 function WinColor      (r,g,b: byte): tColor;                       overload;
procedure Color2RGB     (color: tColor; var r,g,b: byte);            overload;
 function SetAlpha      (color: tColor32; cInt: byte): tColor32;     overload;
 function ShadeColor    (color: tColor; cInt: byte): tColor;         overload;
 function MixColor      (c1,c2: tColor; cInt: byte = 127): tColor;   overload;
 function MixHalfColor  (c1,c2: tColor): tColor;                     overload;
 function GrayValue     (color: tColor): byte;                       overload;
 function RandomColor   (min, max: byte): tColor;                    overload;
 function RandomColor   (min, max: tColor): tColor;                  overload;
 function ColorString      (color: tColor): string;                  overload;
 function ColorShortString (color: tColor): string;                  overload;
 function ColorHexString   (color: tColor): string;                  overload;

//=== t C o l o r 1 5 ===
 function RGB15 (r,g,b: byte): word;

//=== t C o l o r 1 6 ===
 function RGB16 (r,g,b: byte): word;

//=== t C o l o r 3 2 ===
{$ifndef graph32}
 function Color32 (r,g,b: byte; alpha: byte=255): tColor32;          overload;
{$endif}
 function Color32 (color: tColor32; alpha: byte): tColor32;          overload;
 function RandomColor32 (min, max: byte): tColor32;                  overload;
 function RandomColor32 (min, max: tColor32): tColor32;              overload;
                                       // like windows.TRGBQuad
// function MixColor (c1,c2: tColorRGBI; i2: byte): tColorRGBI;        overload;

 function WinColor (color: tColor32): tColor;                        overload;
 function ShadeColor(color: tColor32; cInt: byte): tColor32;         overload;
 function MixColor (c1,c2: tColor32; i1: byte = 127): tColor32;      overload;
 function MixHalfColor (c1,c2: tColor32): tColor32;                  overload;
 function GrayValue (color: tColor32): tColor32;                     overload;
 function ColorMin (c1,c2: tColor32): tColor32;
 function ColorMax (c1,c2: tColor32): tColor32;
 function ColorAdd (c1,c2: TColor32): TColor32;
 function ColorString      (color: tColor32): string;                overload;
 function ColorShortString (color: tColor32): string;                overload;
 function ColorHexString   (color: tColor32): string;                overload;

{ additional colors:   a r g b}

const
      DarkGray     = gray25;
      DarkRed      = $FF800000;
      LightRed     = $FFFF8080;
      DarkBlue     = $FF000080;
      DarkGreen0   = $FF008000;
      LightGreen   = $FF80FF80;
      LightMagenta = $FFFF80FF;
      DarkBrown    = $FF774422;
      Copper       = $FFB87333;             
      Silver       = $FFE6E8FA;
      WineRed      = $FF990033;
      LightBlue0   = $FF8080FF;
      Orange0      = $FFFF5222;

const Invisible    = $00000000;  // 'nothing do draw'
      halfTransparent = $7FFFFFFF;

//=== t L o n g C o l o r ===
 function LongColor (color: tColor): tLongColor;
 function Color32   (color: tLongColor): tColor32;                   overload;
 function WinColor  (color: tLongColor): tColor;                     overload;
 function ShadeColor(color: tLongColor; cInt: byte): tLongColor;     overload;
procedure IncColor  (var c1,c2: tLongColor);                         overload;
 function MixHalfColor (c1,c2: tLongColor): tLongColor;              overload;
 function ColorString (color: tLongColor): string;                   overload;

//=== t R e a l C o l o r ===
function Color32 (const rgbReal: tRealColor): tColor32;              overload;
function RealColor (const r,g,b: float): tRealColor;                 overload;
function RealColor (const color: tColor): tRealColor;                overload;
function RealColor (const color: tColor32): tRealColor;              overload;
procedure RGBRealGray   (var rgbReal: tRealColor);
procedure RGBNorm       (var rgbReal: tRealColor);
procedure RGBRandom     (var rgbReal: tRealColor);
 function ColorString (color: tRealColor): string;                   overload;

//*********************************************************
IMPLEMENTATION   uses SysUtils;

//===============
//  t C o l o r
//===============

//*********************************************************
// convert r g b values to tColor
// input:     r,g,b:   rgb values (0..255)
// return:    dWord:   24 bit color
// format:    24bit (888) 00000000 bbbbbbbb gggggggg rrrrrrrr
//---------------------------------------------------------
function WinColor (r,g,b: byte): tColor;
begin
  WinColor := r + g shl 8 + b shl 16;
end;
//*********************************************************
// convert color to rgb values
// input:     color:  mode specific color value
// output:    r,g,b:  rgb values of color (0..255)
//---------------------------------------------------------
procedure Color2RGB (color: tColor; var r,g,b: byte);
begin
  r :=  longint(color)         and $FF;
  g := (longint(color) shr  8) and $FF;
  b := (longint(color) shr 16) and $FF;
end;
//*********************************************************
// calculate intensity color
//            ShadeColor = (color * int) / 255
// input:     color:  base colors
//            cInt:   color intensity (0..255)
// return:    tColor: intensity color
//---------------------------------------------------------
function ShadeColor (color: tColor; cInt: byte): tColor; overload;
begin
{$Q-}
  Result := ((longWord(color) and $00FF00FF) * cInt) shr 8 and $FF00FF
           +((longWord(color) and $0000FF00) * cInt) shr 8 and $FF00;
end;
//*********************************************************
// mix 2 colors with given intensity
//            MixColor = (c1 * (255-i1) + c2 * i1) / 256
// input:     c1,c2:  colors to mix
//            cInt:   intensity of color c1 (0..255)
// return:    tColor: mixed color
//---------------------------------------------------------
function MixColor (c1,c2: tColor; cInt: byte = 127): tColor;
var  i2: longWord;
begin
  i2 := 256 - cInt;
  Result := ((longWord(c1) and $00FF00FF) * i2
            +(longWord(c2) and $00FF00FF) * cInt) shr 8 and $FF00FF
           +((longWord(c1) and $0000FF00) * i2
            +(longWord(c2) and $0000FF00) * cInt) shr 8 and $FF00;
end;
//*********************************************************
// mix 2 colors with half intensity together
// note: this function is faster than MixColor(c1,c2,127)
//            MixHalfColor := (c1 + c2) / 2
// input:     c1,c2:  colors to mix
// return:    tColor: mixed color
//---------------------------------------------------------
function MixHalfColor (c1,c2: tColor): tColor;
begin
  Result := ((c1 shr 1) and $7F7F7F7F
            +(c2 shr 1) and $7F7F7F7F);
end;
//*********************************************************
// return gray value (0..255) of given color
//---------------------------------------------------------
function GrayValue (color: tColor): byte;
begin
  with tColorRGBI(color)
  do Result := (77 * r + 150 * g + 29 * b) shr 8;
end;
//*********************************************************
// return a random rgb color with values between min and max
//---------------------------------------------------------
function RandomColor (min, max: byte): tColor;
var rng: word;
begin
  rng := max - min + 1;
  Result := WinColor (min + random(rng)
                     ,min + random(rng)
                     ,min + random(rng));
end;
//*********************************************************
// return a random rgb color within the limits of a rgb cube
// given by min color and max color
//---------------------------------------------------------
function RandomColor (min, max: tColor): tColor;
begin
  with tColorRGBI(min) do
  Result := WinColor (r + random(tColorRGBI(max).r - r + 1)
                     ,g + random(tColorRGBI(max).g - g + 1)
                     ,b + random(tColorRGBI(max).b - b + 1));
end;
//*********************************************************
// convert color values to string
//---------------------------------------------------------
function ColorString (color: tColor): string;
begin
  with tColorRGBI(color)
  do Result := Format('r=%3d  g=%3d  b=%3d',[r,g,b]);
end;
//*********************************************************
// convert color values to short string
//---------------------------------------------------------
function ColorShortString (color: tColor): string;
begin
  with tColorRGBI(color)
  do Result := Format('r%dg%db%d',[r,g,b]);
end;
//*********************************************************
// convert color values to hex string
//---------------------------------------------------------
function ColorHexString (color: tColor): string;
begin
  with tColorRGBI(color)
  do Result := '$' + HexString (r) + ','
                   + HexString (g) + ','
                   + HexString (b);
end;

//===================
//  t C o l o r 1 5
//===================

{*********************************************************}
{ convert r g b values to word color
{ input:     r,g,b:   rgb values (0..255)
{ return:    word:    15 bit color
{ format:    15bit (555) 0rrrrrgg gggbbbbb
{---------------------------------------------------------}
function RGB15 (r,g,b: byte): word;
begin
  RGB15 := b shr 3 or (g shl 2) and $03E0
                   or (r shl 7) and $FC00;
end;

//===================
//  t C o l o r 1 6
//===================

{*********************************************************}
{ convert r g b values to word color
{ input:     r,g,b:   rgb values (0..255)
{ return:    word:    16 bit color
{ format:    16bit (565) rrrrrggg gggbbbbb
{---------------------------------------------------------}
function RGB16 (r,g,b: byte): word;
begin
  RGB16 := b shr 3 or (g shl 3) and $07E0
                   or (r shl 8) and $F800;
end;

//===================
//  t C o l o r 3 2
//===================

//---------------------------------------------------------
// create a new color with no transparency (alpha = 255)
// input:     r,g,b: byte;      red, green, blue
//---------------------------------------------------------
(*
function Color32 (r,g,b: Byte): tColor32;
begin
  Result := $FF000000 + B shl 16 + G shl 8 + R;
 // Result := (((((A shl 8) or B) shl 8) or G) shl 8) or R;
end;
*)
//---------------------------------------------------------
// create a new color with optional transparency
// note:  this conversion is also realized as Gr32.Color32
// input:     r,g,b: byte;      red, green, blue
//            alpha: byte;      transparency
//                              0=invisible, 255=opaque
//---------------------------------------------------------
{ifndef graph32}
function Color32 (r,g,b: byte; alpha: byte=255): tColor32;   overload;
asm
  MOV  AH,alpha
  SHL  EAX,16
  MOV  AH,DL
  MOV  AL,CL
end;
(*
  begin
    Result := alpha shl 24 + b shl 16 + g shl 8 + r;
  end;
*)
{endif}
//*********************************************************
// return color with given transparency (see also Gr32.SetAlpha)
// input:     color: tColor32;
//            alpha: byte;      0=invisible, 255=opaque
// return:    tColor32          new color
//---------------------------------------------------------
function Color32 (color: tColor32; alpha: byte): tColor32;
//  Result := (color and rgbMask) or (alpha) shl 24;
asm
  And eax, rgbMask;  // RGB
  Shl edx, 24;       // alpha
  Or  eax, edx;      // alpha + rgb
end;
//*********************************************************
// return a random color with rgb values between min and max
//---------------------------------------------------------
function RandomColor32 (min, max: byte): tColor32;
begin
  Result :=  alphaMask or word(random2(min,max)) shl 16
                       or word(random2(min,max)) shl 8
                       or word(random2(min,max));
end;
//*********************************************************
// return a random rgb color within the limits of a rgb cube
// given by min color and max color
//---------------------------------------------------------
function RandomColor32 (min, max: tColor32): tColor32;
{$ifdef graph32}
begin
  Result := Gr32.Color32(random2(tColorBGRA(min).r, tColorBGRA(max).r)
                        ,random2(tColorBGRA(min).g, tColorBGRA(max).g)
                        ,random2(tColorBGRA(min).b, tColorBGRA(max).b)
                        ,random2(tColorBGRA(min).a, tColorBGRA(max).a));
end;
{$else}
var rr,bb,gg,aa: byte;
begin
  rr := byte(tColorRGBI(max).r - tColorRGBI(min).r);
  gg := byte(tColorRGBI(max).g - tColorRGBI(min).g);
  bb := byte(tColorRGBI(max).b - tColorRGBI(min).b);
  aa := byte(tColorRGBI(max).a - tColorRGBI(min).a);

  with tColorRGBI(min) do
  Result := Color32((b + random(bb)), (g + random(gg)),
                    (r + random(rr)), (a + random(aa)));
end;
{$endif}
//*********************************************************
// convert tColor32 to windows color (see GR32...)
//---------------------------------------------------------
function WinColor (color: tColor32): tColor;
asm
  // the alpha channel byte is set to zero!
  SHL    EAX,8  // ABGR  ->  BGR0
  BSWAP  EAX    // BGR0  ->  0RGB
end;
//*********************************************************
// mix 2 colors with given intensity (alpha also mixed)
//            MixColor = (c1 * (255-i1) + c2 * i1) / 256
// note:      see also gr32.CombineReg
// input:     c1,c2:  colors to mix
//            i1:     intensity of color c1 (0..255)
// return:    tColor: mixed color
//---------------------------------------------------------
function MixColor (c1,c2: tColor32; i1: byte = 127): tColor32;
var  i2: word;
begin
//  i1 := ((255-(c1 shr 24)) * i1) shr 8;
  i2 := 256 - i1;
  Result := ((c1 and $00FF00FF) * i2
            +(c2 and $00FF00FF) * i1) shr 8 and $00FF00FF
          +(((c2 and $FF00FF00) shr 8 * i1)
           +((c1 and $FF00FF00) shr 8 * i2)) and $FF00FF00;
end;
//*********************************************************
// Set alpha value of color (like Gr32.SetAlpha but with byte value)
// input:     color:  base colors
//            cInt:   color intensity (0..255) 0=transparent
// return:    tColor: intensity color
//---------------------------------------------------------
function SetAlpha (color: tColor32; cInt: byte): tColor32;
begin
  Result := (Color and $00FFFFFF) or (longWord(cInt) shl 24);
end;
//*********************************************************
// calculate shaded color, similar to Gr32.SetAlpha but instead
// of changing alpha value, RGB values will be changed
//            ShadeColor = (color * int) / 255
// input:     color:  base colors
//            cInt:   color intensity (0..255)
// return:    tColor: intensity color
//---------------------------------------------------------
function ShadeColor (color: tColor32; cInt: byte): tColor32;
begin
  Result := ((longWord(color) and $00FF00FF) * cInt) shr 8 and $FF00FF
           +((longWord(color) and $0000FF00) * cInt) shr 8 and $FF00
           +(color and $FF000000);
end;
//*********************************************************
// mix 2 colors with half intensity together
// note: this function is faster than MixColor(c1,c2,127)
//            MixHalfColor := (c1 + c2) / 2
// input:     c1,c2:  colors to mix
// return:    tColor32: mixed color
//---------------------------------------------------------
function MixHalfColor (c1,c2: tColor32): tColor32;
begin
  Result := ((c1 shr 1) and $7F7F7F7F
            +(c2 shr 1) and $7F7F7F7F);
end;
//*********************************************************
// return gray value of given color
//---------------------------------------------------------
function GrayValue (color: tColor32): tColor32;
var   gray: byte;
begin
  with tColorRGBI(color) do begin
    gray := (77 * r + 150 * g + 29 * b) shr 8;
    result := a shl 24 + gray shl 16 + gray shl 8 + gray;
  end;
end;
//*********************************************************
// return color with minimum rgb values
//---------------------------------------------------------
function ColorMin (c1,c2: tColor32): tColor32;
begin
  Result := Min(c1 and alphaMask, C2 and alphaMask)
          + Min(c1 and redMask,   C2 and redMask)
          + Min(c1 and greenMask, C2 and greenMask)
          + Min(c1 and blueMask,  C2 and blueMask);
end;
//*********************************************************
// return color with maximum rgb values
//---------------------------------------------------------
function xColorMax (c1,c2: tColor32): tColor32;
begin
  Result := Max(c1 and alphaMask, C2 and alphaMask)
          + Max(c1 and redMask,   C2 and redMask)
          + Max(c1 and greenMask, C2 and greenMask)
          + Max(c1 and blueMask,  C2 and blueMask);
end;

function ColorMax(C1, C2: TColor32): TColor32;
var
  r1, g1, b1, a1: longword;
  r2, g2, b2, a2: TColor32;
begin
  a1 := C1 and $FF000000;
  r1 := C1 and $00FF0000;
  g1 := C1 and $0000FF00;
  b1 := C1 and $000000FF;

  a2 := C2 and $FF000000;
  r2 := C2 and $00FF0000;
  g2 := C2 and $0000FF00;
  b2 := C2 and $000000FF;

  if a2 > a1 then a1 := a2;
  if r2 > r1 then r1 := r2;
  if g2 > g1 then g1 := g2;
  if b2 > b1 then b1 := b2;

  Result := a1 + r1 + g1 + b1;
end;

//*********************************************************
function ColorAdd (c1,c2: TColor32): TColor32;
var
  r1, g1, b1, a1: Integer;
begin
  a1 := C1 shr 24 + C2 shr 24;
  r1 := C1 and $00FF0000 + C2 and $00FF0000;;
  g1 := C1 and $0000FF00 + C2 and $0000FF00;
  b1 := C1 and $000000FF + C2 and $000000FF;

  if a1 > $FF then a1 := $FF;
  if r1 > $FF0000 then r1 := $FF0000;
  if g1 > $FF00 then g1 := $FF00;
  if b1 > $FF then b1 := $FF;

  Result := tColor32(a1 shl 24 + r1 + g1 + b1);
end;

//*********************************************************
// convert color values to string 'r=...  g=...  b=...  alpha=...'
//---------------------------------------------------------
function ColorString (color: tColor32): string;
begin
  with tColorRGBI(color)
  do Result := Format('r=%3d  g=%3d  b=%3d  alpha=%3d',[r,g,b,a]);
end;
//*********************************************************
// convert color values to short string 'r...g...b...a...'
//---------------------------------------------------------
function ColorShortString (color: tColor32): string;
begin
  with tColorRGBI(color)
  do Result := Format('r%dg%db%da%d',[r,g,b,a]);
end;
//*********************************************************
// convert color values to hex string '$xxxxxxxx'
//---------------------------------------------------------
function ColorHexString (color: tColor32): string;
begin
  Result := '$' + HexString (color);
end;

//*********************************************************
// mix 2 colors with given intensity
//            MixColor = (c1 * (255-i1) + c2 * i1) / 256
// Rearranged formula saves 1 multiplication:
//            MixColor := c1 + (c2-c1) * i1 / 256
// input:     c1,c2:  colors to mix
//            i1:     intensity of color c1 (0..255)
// return:    tColor: mixed color
//---------------------------------------------------------
(*
function MixColor (c1,c2: tColorRGBI; i2: byte): tColorRGBI;
var  i1: longint;
{
begin
  i1 := 255 - i2;
  MixColor.r := (c1.r * i2 + c2.r * i1) shr 8;
  MixColor.g := (c1.g * i2 + c2.g * i1) shr 8;
  MixColor.b := (c1.b * i2 + c2.b * i1) shr 8;
end;
}
begin
  // 2nd formula
  MixColor.b := c1.b + ((longint(c2.b) - c1.b) * i1) div 256;
  MixColor.g := c1.g + ((longint(c2.g) - c1.g) * i1) div 256;
  MixColor.r := c1.r + ((longint(c2.r) - c1.r) * i1) div 256;
end;
*)


//=======================
//  t L o n g C o l o r
//=======================

//*********************************************************
// convert color from byte to 24.8 fix point values
//---------------------------------------------------------
function LongColor (color: tColor): tLongColor;
begin
  with tColorRGBI(color)
  do begin
    result.r := r shl 8;
    result.g := g shl 8;
    result.b := b shl 8;
  end;
end;
//*********************************************************
// convert long color to tColor32
//---------------------------------------------------------
function Color32 (color: tLongColor): tColor32;
begin
  result := color.r shl 16 and redMask
           +color.g        and greenMask
           +color.b shr 8  and blueMask;
end;
//*********************************************************
// convert color from 24.8 fix point to windows color
//---------------------------------------------------------
function WinColor (color: tLongColor): tColor;
begin
  result := color.r shl 8 and redMask
           +color.g       and greenMask
           +color.b shr 8 and blueMask;
end;
//*********************************************************
// calculate intensity color
//            ShadeColor = (color * int) / 255
// input:     color:  base colors
//            cInt:   color intensity (0..255)
// return:    tColor: intensity color
//---------------------------------------------------------
function ShadeColor (color: tLongColor; cInt: byte): tLongColor; overload;
begin
  Result.r := (color.r * cInt) shr 8;
  Result.g := (color.g * cInt) shr 8;
  Result.b := (color.b * cInt) shr 8;
end;
//---------------------------------------------------------
procedure IncColor  (var c1,c2: tLongColor);  //c1 = c1 + c2;
begin
//  c1.r := c1.r + c2.r;
//  c1.g := c1.g + c2.g;
//  c1.b := c1.b + c2.b;
  inc (c1.r, c2.r);
  inc (c1.g, c2.g);
  inc (c1.b, c2.b);
end;
//---------------------------------------------------------
function MixHalfColor (c1,c2: tLongColor): tLongColor;
begin
  Result.r := (c1.r + c2.r) shr 1;
  Result.g := (c1.g + c2.g) shr 1;
  Result.b := (c1.b + c2.b) shr 1;
end;
//---------------------------------------------------------
function ColorString (color: tLongColor): string;
begin
  with color
  do Result := Format('r=%5.3f  g=%5.3f  b=%5.3f',[r / 256, g / 256, b / 256]);
end;

//=======================
//  t R e a l C o l o r
//=======================

(*
//=== t R G B ===                      {used for palette handling}

type  pRGB = ^tRGB;
      tRGB = packed record             {color with longint values}
        rb,gb,bb: byte;   end;         {  rgb range: 0..255}

//*********************************************************
// convert color from tRealColor to tRGB format
//---------------------------------------------------------
procedure RGBRealToByte (var rgbReal: tRealColor;
                         var rgb: tRGB);
begin
  with rgbReal, rgb
  do begin
    rb := byte (trunc (r * maxRealByte));
    gb := byte (trunc (g * maxRealByte));
    bb := byte (trunc (b * maxRealByte));
  end;
end;
//*********************************************************
// convert color from tRGB to tRealColor format
//---------------------------------------------------------
procedure RGBByteToReal (var rgb: tRGB;
                         var rgbReal: tRealColor);
begin
  with rgbReal, rgb
  do begin
    r := rb / maxColorRGB;
    g := gb / maxColorRGB;
    b := bb / maxColorRGB;
  end;
end;
*)
//*********************************************************
// convert color from tRealColor to tColor32 format
//---------------------------------------------------------
function Color32 (const rgbReal: tRealColor): tColor32;
begin                                                 //todo: ??? ungetestet
  Result := byte (trunc (rgbReal.r * maxRealByte))
          +(byte (trunc (rgbReal.g * maxRealByte)) shl 8)
          +(byte (trunc (rgbReal.b * maxRealByte)) shl 16)
          + $FF000000;
end;
//*********************************************************
// convert color values to tRealColor format
// input:      r,g,b     color values (range: 0.0 .. 1.0)
//---------------------------------------------------------
function RealColor (const r,g,b: float): tRealColor;
begin
  Result.r := r;
  Result.g := g;
  Result.b := b;
  RGBNorm (Result);
end;
//*********************************************************
// convert color from tColor to tRealColor format
//---------------------------------------------------------
function RealColor (const color: tColor): tRealColor;
begin
  Result.r := tColorRGBI(color).r / 255.0;
  Result.g := tColorRGBI(color).g / 255.0;
  Result.b := tColorRGBI(color).b / 255.0;
end;
//*********************************************************
// convert color from tColor32 format to tRealColor format
//---------------------------------------------------------
function RealColor (const color: tColor32): tRealColor;

var   alpha: float;
begin                                       // 65280 = 255 * 256
  with Result do begin
    alpha := (color and $FF000000) shr 24;
    r := (color and $00FF0000) shr 16 * alpha / 65280;
    g := (color and $0000FF00) shr  8 * alpha / 65280;
    b := (color and $000000FF)        * alpha / 65280;
  end;
end;
//*********************************************************
// convert rgb values to gray scaled color
//---------------------------------------------------------
procedure RGBRealGray (var rgbReal: tRealColor);

begin
  with rgbReal
  do begin
    r := 0.299 * r + 0.587 * g + 0.114 * b;
    g := r;
    b := r;
  end;
end;
//*********************************************************
// normalize RGB values to range 0..1
//---------------------------------------------------------
procedure RGBNorm (var rgbReal: tRealColor);

var  max: float;

begin
  with rgbReal
  do begin
    if r >= g
    then
      if r >= b
      then max := r
      else max := b
    else   {g > r}
      if g >= b
      then max := g
      else max := b;

    if max > 1.00
    then begin
      r := r / max;
      g := g / max;
      b := b / max;
    end;
  end;
end;
//*********************************************************
// create random color
//---------------------------------------------------------
procedure RGBRandom (var rgbReal: tRealColor);
begin
  rgbReal.r := random(256) / 255.0;
  rgbReal.g := random(256) / 255.0;
  rgbReal.b := random(256) / 255.0;
end;
//*********************************************************
function ColorString (color: tRealColor): string;
begin
  with color
  do Result := Format('r=%5.3f  g=%5.3f  b=%5.3f',[r, g, b]);
end;
{$ifopt D+}
//*********************************************************
procedure Test1;
begin
  writeln (tst, ColorString (Color32(redRealColor)));
  writeln (tst, ColorString (Color32(greenRealColor)));
  writeln (tst, ColorString (Color32(blueRealColor)));
  writeln (tst, ColorString (Color32(whiteRealColor)));
end;
//*********************************************************
procedure Test2;
const rgbiColor: tColorRGBI = (r:11; g:22; b:33; a:44);
const longColor: tLongColor = (r:-maxInt; g:maxInt; b:0);
//const longColor: tLongColor = (r:256; g:128; b:0);
const realColor: tRealColor = (r:1.0; g:0.5; b:0.2);

begin
  writeln (tst, ColorString (tColor32(rgbiColor)));
  writeln (tst, ColorString (longColor));
  writeln (tst, ColorString (realColor));
end;
//*********************************************************
//begin
//  Test1;
//  Test2;
{$endif}

end.

