
//*********************************************************
//
//  UNIT:     S _ S C R E E N     01.11.2003,  12.09.2005
//
//  ScreenSaver drawing window handling:
//  handle ScreenSaver activation and deactivation
//  and draw test objects
//
//*********************************************************
unit S_SCREEN;

interface

uses
  SysUtils, Windows, Messages, Graphics,
  Forms, Dialogs, Classes, ExtCtrls, Controls,
  GR32, GR32_Polygons, GR32_Image;

type
  tScreenForm = class(TForm)
    timer1: TTimer;
    Image32: TImage32;
    procedure timer1Timer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormHide(Sender: TObject);
    procedure Restart (windowed: boolean);
    procedure TogglePaused;
    procedure SetParentWindow(pwh: HWND);
    procedure FormDestroy(Sender: TObject);
    procedure Clear;
    procedure ShowText (x,y: integer; txt: string;
                        color: tcolor32 = clYellow32);

  private
    { Private declarations }
    procedure HandleMessage(var Msg: TMsg; var Handled: boolean);
  public
    parentWindowHandle: HWND;
    paused: boolean;
    property IsPaused: boolean     read paused;
  end;

var   ScreenForm: tScreenForm;

//*********************************************************
implementation

uses  SysUtil2, S_Config, Gr32_Ext, Gr32_Gouraud;

{$R *.DFM}

var   mousePos:  TPoint  = (x:0; y:0);
      winTest:   boolean = false;
      cursorCnt: integer = 0;

      bm32: tBitmap32 = nil;
      TriaPoly, Outline: tPolygon32;

//---------------------------------------------------------
procedure CursorOn;
begin
  if cursorCnt < 0 then cursorCnt := ShowCursor(true);
end;
//---------------------------------------------------------
procedure CursorOff;
begin
  if cursorCnt >= 0 then cursorCnt := ShowCursor(false);
end;
//---------------------------------------------------------
procedure tScreenForm.FormCreate(Sender: TObject);
begin
  inherited;
  parentWindowHandle := 0;
  bm32 := Image32.Bitmap;
  bm32.Font.Color := clYellow;
  SelectBitmap32 (bm32);

  triaPoly := tPolygon32.Create;
  triaPoly.AntiAliased := true;

  Outline  := tPolygon32.Create;
  MarkTime;
end;
//---------------------------------------------------------
procedure tScreenForm.FormDestroy(Sender: TObject);
begin
  Outline.Free;
  TriaPoly.Free;
end;
//---------------------------------------------------------
procedure tScreenForm.FormShow(Sender: TObject);
begin
  GetCursorPos(mousePos);
  timer1.Enabled        := true;
  Application.OnMessage := HandleMessage;

  if winTest
  then CursorOn
  else CursorOff;

  Image32.Bitmap.SetSize (clientWidth, clientHeight);
end;
//*********************************************************
// Llschen des Zeichenbereiches
//---------------------------------------------------------
procedure tScreenForm.Clear;
begin
  bm32.Clear;
end;
//*********************************************************
// Text an Grafikposition x,y ausgeben
// Es wird die mit SetTextColor gesetzte Farbe verwendet.
//---------------------------------------------------------
procedure tScreenForm.ShowText (x,y: integer; txt: string;
                                color: tcolor32 = clYellow32);
begin
  bm32.Font.Color := WinColor (color);
  bm32.TextOut (x,y, txt);
end;
//---------------------------------------------------------
procedure tScreenForm.FormResize(Sender: TObject);
begin
//  bm32.Clear;     // SetSize lscht Bitmap automatisch
  bm32.SetSize (width, height);

//  ShowText (1, 1, IntToStr(width));
//  ShowText (1,20, IntToStr(height), clWhite32);
end;
//---------------------------------------------------------
procedure tScreenForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case key of
    19: TogglePaused;  // Pause-key
  end;
end;
//---------------------------------------------------------
procedure tScreenForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  case key of
    'p': TogglePaused;  // Pause on/off
    else Close;
  end;
end;
//---------------------------------------------------------
procedure tScreenForm.FormHide(Sender: TObject);
begin
  Application.OnMessage := nil;
  timer1.Enabled        := false;
  CursorOn;
end;
//---------------------------------------------------------
procedure tScreenForm.SetParentWindow (pwh: HWND);
var   rect: tRect;
begin
  if pwh = 0 then exit;

  GetWindowRect (pwh, rect);

  screenForm.parentWindowHandle := Windows.SetParent (handle, pwh);
  with rect do screenForm.SetBounds (0,0, right - left, bottom - top);

  if screenForm.parentWindowHandle = 0
  then Application.MessageBox('SetParentWindow failed'
                             ,'info', MB_OK + MB_DEFBUTTON1);
end;
//---------------------------------------------------------
procedure tScreenForm.HandleMessage(var Msg: TMsg; var Handled: boolean);
var   mouseAct: TPoint;
      whdl: HWND;
begin
  case Msg.message of
    WM_RBUTTONDOWN: begin
      ConfigForm.Show;
      CursorOn;
    end;
  end;
  whdl := GetFocus;
  if whdl <> self.handle then exit;
  if winTest then exit;

  case Msg.message of
    WM_MOUSEMOVE: begin
        if screenForm.parentWindowHandle <> 0    // preview startet ?
        then exit;                               // don't close on mouse move
        GetCursorPos(mouseAct);
        if (Abs(mouseAct.x - mousePos.x) > 10)
        or (Abs(mouseAct.y - mousePos.y) > 10)
        then Close;
      end;
//    WM_ACTIVATE,
//    WM_NCACTIVATE,  WM_ACTIVATEAPP,
    WM_LBUTTONDOWN, WM_MBUTTONDOWN,
    WM_SYSKEYDOWN,  WM_SYSKEYUP:
      Close;
  end;
end;
//---------------------------------------------------------
var   objects:    integer =  1;
      oSize:      integer = 40;
      oAlpha:     byte = 127;
      oThickness: byte =   1;
      oSteps:     byte =   5;
//---------------------------------------------------------
procedure tScreenForm.Restart (windowed: boolean);
begin
  paused := false;
  objects := 0;

  winTest := windowed;
  if windowed
  then begin
    BorderStyle := bsSizeable;
    WindowState := wsNormal;
//    FormStyle := fsStayOnTop;
  end
  else begin
    BorderStyle := bsNone;
    WindowState := wsMaximized;
//    FormStyle := fsNormal;
  end;
  bm32.Clear;
end;
//*********************************************************
// turn pause on / off
//---------------------------------------------------------
procedure tScreenForm.TogglePaused;

begin
  paused := not paused;
  if paused
  then Caption := 'SSDemo2 (paused)'
  else Caption := 'SSDemo2'
end;
//*********************************************************
// return a random color with rgb values between min and max
// and optional alpha value (transparency)
//---------------------------------------------------------
function RandomColor (min, max: byte; int: byte = $FF): tColor32;

var rng: word;    // 0..256 !!!
begin
  rng := max - min + 1;
  Result := Color32(byte(min + random(rng))
                   ,byte(min + random(rng))
                   ,byte(min + random(rng)), int);
end;
//*********************************************************
// return a random color with rgb values between min and max
// and with a random alpha value (transparency)
//---------------------------------------------------------
function RandomAlphaColor (min, max: byte): tColor32;
var rng: word;
begin
  rng := max - min + 1;
  Result := Color32(byte(min + random(rng))
                   ,byte(min + random(rng))
                   ,byte(min + random(rng))
                   ,byte(min + random(rng)));
end;
//*********************************************************
var   maxLines: word = 22;
      r1: word = 20;
      r2: word = 40;

procedure TestLines (x,y: integer; lCol: tColor32);

var   li: integer;   angle, sa,ca: single;
begin
  maxLines := 10 + oSize div 3;
  r1 := oSize div 3;
  r2 := oSize div 2;
  for li := 1 to maxLines
  do begin
    angle := Pi * 2 * li / maxLines;
    sa := sin(angle);
    ca := cos(angle);
    bm32.LineAS (x + round (r1*sa),   y + round (r1*ca)
                ,x + round (2*r2*sa), y + round (r2*ca), lCol, false);
  end;
end;
//---------------------------------------------------------
procedure tScreenForm.timer1Timer(Sender: TObject);

var
  p0,p1,p2   : tPoint;
  a,b,s2     : integer;
  fullcolor, blendColor: tColor32;
  triaCol    : tColor3v; //array[0..2] of tColor32;
  ca1,ca2,ca3,ca4: tColor32;
  triangle   : tTriangle;

  procedure RandomPosition;
  begin
    p0.x := Random(ClientWidth - oSize);
    p0.y := Random(ClientHeight - oSize);
  end;

  procedure RandomSize;
  begin
    s2 := oSize div 2;
    p1.x := Random(ClientWidth);
    p1.y := Random(ClientHeight);
    p0.x := p1.x - s2;
    p0.y := p1.y - s2;
    p2.x := p1.x + s2;
    p2.y := p1.y + s2;
    a := s2;
    b := s2 + Random(s2) - s2 div 2;
  end;

  procedure RandomTriangle;
  var p: tPoint;
  begin
    RandomSize;
    triaPoly.Clear;
    p.x := p1.x + random(oSize);
    p.y := p0.y;
    triaPoly.Add (FixedPoint(p));
    p.x := p0.x;
    p.y := p1.y + random(oSize);
    triaPoly.Add (FixedPoint(p));
    p.x := p2.x;
    p.y := p1.y + random(oSize);
    triaPoly.Add (FixedPoint(p));
  end;

  procedure RandomTriangle2;
  begin
    RandomSize;
    with triangle do begin
      a.x := p1.x + random(oSize);
      a.y := p0.y;
      b.x := p0.x;
      b.y := p1.y + random(oSize);
      c.x := p2.x;
      c.y := p1.y + random(oSize);
    end;
  end;

  // generate 4 random colors
  procedure RandomColors (transparent: boolean);
  begin
    RandomSize;
    if transparent then begin
      ca1 := RandomAlphaColor (10, 255);
      ca2 := RandomAlphaColor (10, 255);
      ca3 := RandomAlphaColor (10, 255);
      ca4 := RandomAlphaColor (10, 255);
    end else begin
      ca1 := RandomColor (10, 255);
      ca2 := RandomColor (10, 255);
      ca3 := RandomColor (10, 255);
      ca4 := RandomColor (10, 255);
    end;
  end;

  procedure RandomTriangleColor (transparent: boolean);
  begin
    RandomTriangle2;
    if transparent then begin
      triaCol.a := RandomAlphaColor (10, 255);
      triaCol.b := RandomAlphaColor (10, 255);
      triaCol.c := RandomAlphaColor (10, 255);
    end else begin
      triaCol.a := RandomColor (10, 255);
      triaCol.b := RandomColor (10, 255);
      triaCol.c := RandomColor (10, 255);
    end;
    tColorRGBA(triaCol.a).a := 255;
    tColorRGBA(triaCol.b).a := 255;
    tColorRGBA(triaCol.c).a := 255;
  end;

  procedure RandomSplineCurves (antiAliasing, filled: boolean;
                                color: TColor32; thickness, steps: byte);
  const pnr = 4;
        closedCurve = true;
  var   ni: integer;
        p2d: TArrayOfPoint;            // single float points
        points: TArrayOfFixedPoint;    // 24.8 fixpoint points
  begin
    SetLength (p2d, pnr);
    SetLength (points, pnr);

    // create random polygon
    RandomPosition;
    for ni := 0 to pnr - 1 do with p2d[ni] do begin
      x := p0.x +random(50+oSize);
      y := p0.y +random(50+oSize);
      points[ni] := FixedPoint(p2d[ni]);
    end;
//    PolylineTS (bm32, points, $FF666666, closedCurve);   // only for testing
    DrawCubicCurve (points, steps, color
                   ,closedCurve, antiAliasing, filled, thickness);
  end;

  procedure TestSplineCurves (antiAliasing, filled: boolean;
                              color: TColor32; thickness, steps: byte);
  var   xp, yp: integer;
  var   points: TArrayOfFixedPoint;    // 24.8 fixpoint points
  begin
    SetLength (points, 4);            // 'C'
    xp := p1.X;
    yp := p1.y;
    points[0] := FixedPoint(xp+40, yp);
    points[1] := FixedPoint(xp   , yp);
    points[2] := FixedPoint(xp   , yp+40);
    points[3] := FixedPoint(xp+40, yp+40);
    DrawCubicCurve (points, steps, color, false, antiAliasing, false, thickness);

    SetLength (points, 4);          // 'O'
    xp := p1.X+50;
    yp := p1.Y;
    points[0] := FixedPoint(xp   , yp);
    points[1] := FixedPoint(xp+40, yp);
    points[2] := FixedPoint(xp+40, yp+40);
    points[3] := FixedPoint(xp   , yp+40);
    DrawCubicCurve (points, steps, color, true, antiAliasing, false, thickness);
  end;

begin {timer1Timer}
  if paused then exit;
  SetMasterAlpha (255);

  oSize :=       configForm.gSize.position;
  oAlpha :=      configForm.gAlpha.position;
  oThickness :=  configForm.gThick.position;
  oSteps :=      configForm.gSteps.position;

  if objects > configForm.gObjects.position
  then begin
    objects := 0;
    bm32.Clear;
    bm32.TextOut (10,10, IntToStr (GetRuntimeMS));
    MarkTime;
  end;
  Inc (objects);

  //--- Anti-Aliasing ---
  fullColor := RandomColor (10,255, 255);

  RandomSize;
  if configForm.iLines.Checked
  then TestLines (p1.x, p1.y, fullColor);

  RandomSize;
  if configForm.iBars.Checked
  then bm32.FillRectS (p0.x, p0.y, p2.x, p2.y, fullColor);

  fullColor := RandomColor (10,255);
  RandomTriangle;
  if configForm.iFillTria.Checked
  then triaPoly.DrawFill (bm32, fullColor);

  fullColor := RandomColor (10,255);
  if configForm.iTriangles.Checked
  then begin
    if oThickness <= 1
    then triaPoly.DrawEdge (bm32, fullColor)
    else begin
      Outline := triaPoly.Outline;
      Outline := Outline.Grow(Fixed(oThickness),1);
      Outline.DrawFill(bm32, fullColor);
    end;
  end;

  if configForm.iSplineCurve.Checked
  then RandomSplineCurves (true, false, fullColor, oThickness, oSteps);

  if configForm.iSplineArea.Checked
  then RandomSplineCurves (true, true, fullColor, oThickness, oSteps);

  if configForm.iEllipse.Checked
  then begin
    RandomColors (false);
    DrawEllipse (p1.x, p1.y, p2.x-p1.x, (p2.y-p1.y) div 2, random(180)
                ,fullcolor, 36, oThickness, true, false);
  end;


  //--- AlphaBlending ---

  blendColor := RandomColor (10,255,oAlpha);
  RandomSize;
  if configForm.bLines.Checked
  then TestLines (p1.x, p1.y, blendColor);

  if configForm.bBars.Checked
  then bm32.FillRectTS (p1.x, p1.y, p2.x, p2.y, blendColor);

  RandomTriangle;                                // build random triangle

  if configForm.bTriangles.Checked               // draw triangle
  then begin
    if oThickness <= 1
    then triaPoly.DrawEdge (bm32, blendColor)
    else begin
      Outline := triaPoly.Outline;
      Outline := Outline.Grow(Fixed(oThickness),1);
      Outline.DrawFill(bm32, blendColor);
    end;
  end;

  if configForm.bFillTria.Checked                // draw filled triangle
  then triaPoly.DrawFill (bm32, blendColor);

  if configForm.bSplineCurve.Checked
  then RandomSplineCurves (true, false, blendColor, oThickness, oSteps);

  if configForm.bSplineArea.Checked
  then RandomSplineCurves (true, true, blendColor, oThickness, oSteps);

//  TestSplineCurves (true, true, blendColor, oThickness, oSteps);

  if configForm.bEllipseDisk.Checked
  then begin
    RandomColors (false);
    DrawEllipse (p1.x, p1.y, p2.x-p1.x, (p2.y-p1.y)div 2, random(180)
                ,blendcolor, 27, 1, true, true);
  end;

  //--- Gouraud drawing ---
  SetMasterAlpha (255);
  if configForm.gBars.Checked
  then begin
    RandomColors (false);
    GouraudRectangle (p0.x, p0.y, p2.x, p2.y, ca1,ca2,ca3,ca4);
  end;

  if configForm.gTriangles.Checked
  then begin
    RandomTriangleColor (false);
    GouraudTriangle (triangle, triaCol);
  end;

  //--- Gouraud Blending ---

  SetMasterAlpha (oAlpha);

  if configForm.bgBars.Checked
  then begin
    RandomColors (true);
    GouraudRectangle (p0.x, p0.y, p2.x, p2.y, ca1,ca2,ca3,ca4);
//    DrawRGBColors (p0.x, p0.y, p2.x-p0.x, p2.y-p0.y, 255);
  end;

  if configForm.bgTriangles.Checked
  then begin
    RandomTriangleColor (true);
    GouraudTriangle (triangle, triaCol);
  end;

  activeBitmap.Changed;
end;

end.

