
//*******************************************************************
//
//  UNIT:     S y s U t i l 2               14.01.2000,  26.07.2006
//
//  This Unit consists of additional system routines for Delphi-5/6/7/8
//  and it's Freeware
//
//-------------------------------------------------------------------
//  Min                      minimum value
//  Max                      maximum value
//  Clamp                    return value between given limits
//  ClampByte                return value between 0 and 255
//  ClampWord                return value between 0 and 65535
//  Swap                     swap values
//  Sort                     sort values
//  SAR_?                    arithmetic shift right
//  SqrInt                   square of integer value
//  MulDiv                   computes a*b/c with integers
//  PointerAdd               pointer value addition
//  PointerDiff              pointer value difference
//
//  ZeroMem                  fast set memory to zero
//  SwapMem                  fast swap memory values
//  FillLongWord             fill longword memory
//  MoveLongWord             copy longword memory
//
//  RightPos                 search string from right to left
//  GetLeftString            return searched left string zurck
//  GetRightString           return searched right string zurck
//  TrimAll                  trim string
//  IsInteger                check if string is a valid longint value
//  IsInt64                  check if string is a valid int64 value
//  IsFloat                  check if string is a valid float value
//
//  BinString                convert value -> binary string
//  HexString                convert value -> hexadecimal string
//  HexToLong                convert hexadecimal string -> int64
//  IntString                convert integer value -> string
//  IntDotString             convert integer value -> string with dots
//  FloatString              convert float value -> string
//  StrToFloat2              convert float string -> float value
//  TrueFalse                convert boolean -> 'true' or 'false' string
//  OnOff                    convert boolean -> 'on' or 'off' string
//  YesNo                    convert boolean -> 'yes' or 'no' string
//
//  Sound                    play sound with given frequency
//  NoSound                  turn off sound
//  PlayWaveSound            play file wave sound at background
//  SyncWaveSound            play file wave sound and wait
//  StopWaveSound            stop playing wave sound
//  SetWaveSound             wave sound on/off
//  WaveSoundVolume          set wave sound volume to 0..100 %
//  SetWavePlaybackRate      set wave out pitch
//  ShowWaveError            display last wave out error
//  ErrorBeep                play error sound
//  WarningBeep              play warning sound
//
//  GetTime                  return time values
//  GetDate                  return date values
//  SystemTimeString         return actual system time as string [hh:mm:ss]
//  SystemDateString         return actual system date as string [mm-dd-yyyy]
//  SystemDatumString        return actual system date as string [dd.mm.yyyy]
//  MarkTime                 set time marker
//  GetRunTimeMS             return runtime since last MarkTime in milliseconds
//  PictureRuntimeMS         return drawing time of one picture in milliseconds
//  PictureRuntime           return drawing time of one picture in seconds
//  PicturesPerSecond        return pictures (frames) per second
//  FPS                      return frames per second
//
//  DelphiIsRunning          return true if Delphi IDE is running
//  DelphiVersion            return Delphi development version
//  OperatingSystem          return windows operating system string
//  WindowsVersion           return windows version string
//  GetComputername          return name of computer
//  GetProgramName           return own program name
//  CurrentUserName          return name of current user
//  OSLanguage               return operating system language
//  DirectXVersion           return DirectX version
//  DefaultPrinterName       return default printer name
//
//  CPUSpeed                 return CPU speed [MHz]
//  CPUSpeed2                return CPU speed [MHz] (with priority handling)
//  ProcessorType            return type of processor
//  ProcessorName            return name of processor
//
//  FreeAndNil               free object memory and set object reference to Nil 
//  MemoryStatus             return memory values
//  TotalPhysicalMemory      return whole physical memory in bytes
//  AvailablePhysicalMemory  return available physical memory in bytes
//  TotalVirtualMemory       return whole virtual memory in bytes
//  AvailableVirtualMemory   return available virtual memory in bytes
//
//  GetDisplayInfo           return screen display information
//  DisplayInfoString        return screen display information string
//  ResetDisplayMode         reset display mode
//  SetDisplayMode           set display mode to given display mode index
//  ChangeDisplayMode        set display mode and display result message
//  GetDisplayModeList       get string list of all display modes
//
//  ShowTaskBar              show or hide system task bar
//
//  GetDriveList             return list of existing drives
//  DriveReady               return drive state and sizes
//  GetVolumeName            return volume name of given drive
//  GetCDList                return list of CD/DVD drives
//
//  RestoreStartDirectory    set current directory to start directory
//  GetFileList              return file list
//  GetFileSize              return file size of given file in bytes
//  FileSizeByName           return file size of given file in bytes
//  IsFileWriteable          check, if file can be written
//  SubstituteFilename       return substitute filename
//  GetFilenameExtension     return filename extension after last dot
//  RemoveFilenameExt        remove filename extension with the dot
//  CopyFile                 copy a file using windows function
//
//  InitMutex                create mutex (mutual exclusion)
//  ClearMutex               clear mutex at program end
//  StopMultiStartedProgram  stop double startet program
//  DOSCommand               execute DOS command
//  RunApplication           execute Windows application
//  RunControlPanelApplet    execute system control panel applet
//  RunLink                  execute internet link
//
//  RaiseObjectNilError      raise error object = nil
//  RaiseObjectInitError     raise error object not initialized
//  RaiseMemoryError         raise heap memory error
//  RaiseProcedureError      raise procedure address error
//
//  ShowMessageBox           display message box
//  ShowHint                 display hint [OK]
//  ShowWarning              display warning [OK] [cancel]
//  ShowError                display error  [cancel] [repeat] [ignore]
//  ShowFatalError           display fatal error and quit program
//  ShowFileError            display file error
//
//  -- stack routines ---
//  PushInteger         copy long value onto stack
//  PopInteger          remove long value from stack
//  PushBoolean         copy boolean value onto stack
//  PopBoolean          remove boolean value from stack
//  PushFloat           copy float value onto stack
//  PopFloat            remove float value from stack
//
//  -- classes ---
//  TObjectStack     replacement of standard unit Contnrs.TObjectStack (faster)
//    Create            create a stack object and initialize data
//    Destroy           kill stack object (attention: call Free only!)
//    Count             return number if list elements
//    Push              put copy of element reference onto stack
//    Pop               remove top element reference from stack and return it
//    Peek              return reference to top element of stack
//    Clear             clear all stack elements
//    AtLeast           check if stack has at least given number of elements
//
//  --- global variables ---
//  programName: string      name of ownn program
//  computerName: string     name of own computer
//  ioError: Integer         last file error
//  memStat: TMemoryStatus   memory status
//  msgBoxIsOpen: boolean    true, if modal message box is displayed
//  picCnt: longWord         picture counter
//  screenInfo: tDeviceInfo  last screen device information
//  startScreenInfo: -"-     screen device information at program start
//  startDisplayMode         display mode at program start
//  startDirectory: string   actual directory path at program start
//  tst: text;               debug file, if compiler option is $D+
//
//*******************************************************************
//  Diese Unit besteht aus zustzlichen Routinen fr Delphi-5/6/7/8
//  und sie ist Freeware.
//
//  -- Stack Routinen ---
//  PushInteger         Long-Wert auf den Stapel kopieren
//  PopInteger          Long-Wert vom Stapel nehmen
//  PushBoolean         Boolean-Wert auf den Stapel kopieren
//  PopBoolean          Boolean-Wert vom Stapel nehmen
//  PushFloat           Float-Wert auf den Stapel kopieren
//  PopFloat            Float-Wert vom Stapel nehmen
//
//  -- Klassen ---
//  TObjectStack     Ersatz der Standardklasse Contnrs.TObjectStack (schneller)
//    Create            Erzeugt Stack-Objekt und initialisiert dessen Daten
//    Destroy           Gibt Stack frei (Achtung: nur Free aufrufen!)
//    Count             Gibt die Anzahl der Listelemente zurck
//    Push              Kopie einer Elementreferenz auf dem Stapel kopieren
//    Pop               Entfernt oberste Elementreferenz vom Stapel und gibt diese zurck
//    Peek              Gibt Referenz auf die oberste Adresse des Stapels zurck
//    Clear             Lscht alle Elemente in die Liste
//    AtLeast           Prgt, ob Stack zumindest die angegebene Elementanzahl hat
//
//*******************************************************************
{$I comp_opt.inc}

{$ifdef DEBUG}
  {$define XstackTest  Objektstapelfunktionen}
{$endif}

unit SysUtil2;

{$define useAsm          definition to use asm-code}

interface

uses  SysUtils, Forms, Windows, Classes, Graphics, Types, MMSystem;

{$I float.inc}

const No = False;
      Yes = True;

const NUL = #0;
      BEL = #7;
      DEL = #8;
      TAB = #9;
      LF  = #10;
      FF  = #12;
      CR  = #13;
      ESC = #27;
      BLANC = #32;
      CRLF = #13#10;

const VK_PAGE_UP = 33;    // VK_PRIOR = 33;
      VK_PAGE_DOWN = 34;  // VK_NEXT = 34;

const hexChar: array[0..15] of char = '0123456789ABCDEF';

      // halt error codes
const errZeroDivision      = 200;
      errOutOfRange        = 201;
      errStackOverflow     = 202;
      errHeapOverflow      = 203;
      errNoObjectInit      = 210;     // e.g. wrong object id
      errOutOfMemory       = 217;     // e.g. GetMem
      errNilPointer        = 221;     // e.g. @procCall = nil
      errNilObject         = 222;     // e.g. New
      errObjectType        = 223;
      errNoTypeDefined     = 224;
      errFatalError        = 225;
      errNoGraphicActive   = 230;
      errFatalGraphicError = 231;
      errWrongGraphicMode  = 232;
      errNotImplemented    = 233;
      errUserBreak         = 255;

const maxByte     = $FF;         // 255
      maxWord     = $FFFF;       // 65.535
      maxLongWord = $FFFFFFFF;   // 4.294.967.295;
      maxInt64    = $7FFFFFFFFFFFFFFF;
      // system.maxInt = 32767;
      // system.maxLongInt = 2147483647;

type  //--- static array types with defined length additional to unit SysUtils ---

      PShortArray = ^TShortArray;
      TShortArray = array[0..maxWord] of shortInt;  // 8 Bit

      PSmallIntArray = ^TSmallIntArray;
      TSmallIntArray = array[0..maxWord] of smallInt;  // 16 Bit

      PLongIntArray = ^TLongIntArray;
      TLongIntArray = array[0..maxWord] of longInt;  // 32 Bit

      tFloatArray = array[0..16382] of float;
      pFloatArray = ^tFloatArray;

      //--- dynamic array types additional to unit Types.pas ---
      pShortIntArray   = ^TShortIntDynArray;

      pIntegerArray    = ^TIntegerDynArray;

      pFloatDynArray   = ^tFloatArray;
      tFloatDynArray   = array of float;

      pPointerArray    = ^TPointerArray;
      tPointerArray    = array of pointer;

      pPointArray      = ^tPointArray;      // longint point(x,y)
      tPointArray      = array of Types.TPoint;

      pSmallPointArray = ^tSmallPointArray; // smallint point(x,y)
      tSmallPointArray = array of Types.tSmallPoint;

    //--- hardware information - see GetHardwareInfo ----
    tHardwareInfo = record
      processorType: string;
      processorName: string;
      CPUSpeed: integer;
      mousePresent: boolean;
      mouseButtons: integer;
    end;

    //--- device information - see GetDisplayInfo ----
    tDeviceInfo = record
      dcHandle       : HDC;            // device contents handle
      x, y           : word;           // screen resolution [pixel]
      bitsPerPixel   : word;           // bits per pixel
      pixelFormat    : tPixelFormat;   // pfDevice or bits per pixel
      pPlanes        : word;           // pixel planes
      xSize, ySize   : word;           // screen size [mm]
      xAspect        : word;           // aspect ratio
      yAspect        : word;
      colRes         : word;           // colors = color resolution
      dVersion       : word;           // driver version
      // for Win-NT only
      refreshRate    : word;           // vertical refresh rate [hz]
      xDesktop, yDesktop: word;        // desktop size
    end;

// random functions
function random2 (min,max: integer): integer;                        overload;
function random2 (min,max: float): float;                            overload;

// get minima or maxima (see also unit Math}
// ermittle Minimum und Maximum (siehe auch Unit Math)
function Min (const a,b: longWord): longWord;                        overload;
function Min (const a,b: longInt): longInt;                          overload;
function Min (const a,b: float): float;                              overload;
function Min (const a,b,c: Integer): Integer;                        overload;

function Max (const a,b: longWord): longWord;                        overload;
function Max (const a,b: longInt): longInt;                          overload;
function Max (const a,b: float): float;                              overload;
function Max (const a,b,c: Integer): Integer;                        overload;

// return value between min and max value   // Werterckgabe zw. min und max
function Clamp (const value, min, max: byte): byte;                  overload;
function Clamp (const value, min, max: longint): longint;            overload;
function Clamp (const value, min, max: int64): int64;                overload;
function Clamp (const value, min, max: float): float;                overload;
function ClampByte (value: integer): byte;                           overload;
function ClampByte (value: float): byte;                             overload;
function ClampWord (value: integer): word;

// swap values   // vertausche Werte   bytes:
procedure Swap (var a,b: byte);        {1}                           overload;
procedure Swap (var a,b: shortInt);    {1}                           overload;
procedure Swap (var a,b: word);        {2}                           overload;
procedure Swap (var a,b: smallInt);    {2}                           overload;
procedure Swap (var a,b: longInt);     {4}                           overload;
procedure Swap (var a,b: cardinal);    {4}                           overload;
procedure Swap (var a,b: int64);       {8}                           overload;
procedure Swap (var a,b: pointer);     {4}                           overload;

// sort values    // Sortiere Werte
procedure Sort (var a,b: byte);                                      overload;
procedure Sort (var a,b: word);                                      overload;
procedure Sort (var a,b: smallInt);                                  overload;
procedure Sort (var a,b: longInt);                                   overload;
procedure Sort (var a,b: longWord);                                  overload;
procedure Sort (var a,b: int64);                                     overload;

// rotate enumeration index   // rotiere Aufzhlungsindex
procedure IncIndex (var index: integer; const min,max: integer; delta: integer = 1);
procedure DecIndex (var index: integer; const min,max: integer; delta: integer = -1);

// arithmetic shift operations     // arithmetische Schiebeoperationen
function SAR_2 (value: integer): integer;  assembler;
function SAR_4  (value: integer): integer;
function SAR_8  (value: integer): integer;
function SAR_9  (value: integer): integer;
function SAR_12 (value: integer): integer;
function SAR_13 (value: integer): integer;
function SAR_14 (value: integer): integer;
function SAR_15 (value: integer): integer;
function SAR_16 (value: integer): integer;

// integer arithmetic
function SqrInt (const value: Integer): Integer;
function SqrtInt (const value: Integer): Integer;
function MulDiv (const a, b, c: Integer): Integer;

// pointer arithmetic
function PointerAdd (ptr: pointer; byteSize: integer): pointer;
function PointerDiff (ptr1, ptr2: pointer): integer;

// memory operations      // Speicherbearbeitung
procedure ZeroMem (var dest; sizeof: integer); 
procedure SwapMem (var source, dest; len: integer);
procedure FillLongWord (var dest; count: integer; value: longWord);
procedure MoveLongWord (const source; var dest; count: Integer);

// string handling
function RightPos       (s: string; substr: string): byte;
function GetLeftString  (s: string; substr: string): string;
function GetRightString (s: string; substr: string): string;
function TrimAll        (const s: string): string;                   overload;
function TrimAll        (const s, subStr: string): string;           overload;
//function IsInteger      (s: string): Boolean;
//function IsInt64        (s: string): Boolean;
function IsFloat        (s: string): boolean;

// conversions      // Konvertierung
function BinString (value: byte): string;                            overload;
function BinString (value: word): string;                            overload;
function BinString (value: longint): string;                         overload;
function BinString (value: int64): string;                           overload;
function HexString (value: byte): string;                            overload;
function HexString (value: word): string;                            overload;
function HexString (value: longWord): string;                        overload;
function HexString (value: longWord; sep: string): string;           overload;
function HexString (value: longInt): string;                         overload;
function HexString (value: longInt; sep: string): string;            overload;
function HexString (value: int64): string;                           overload;
function HexString (value: int64; sep: string): string;              overload;
function HexString (value: pointer): string;                         overload;
function HexString (value: tProcedure): string;                      overload;
function HexToLong (text: string): int64;
function IntString    (value: integer; digits: integer = 0): string;
function IntDotString (value: integer; digits: integer = 0): string;
function FloatString  (value: extended; precision: integer = 0;
                                        digits: integer = 0): string;
function StrToFloat2  (fs: string): Extended;
function TrueFalse    (flag: boolean): string;
function OnOff        (flag: boolean): string;
function YesNo        (flag: boolean): string;

// sound handling
procedure Sound (Hz: word);
procedure NoSound;

procedure NameWaveSound  (soundFile: string);
procedure LoopWaveSound (soundFile: string);
procedure PlayWaveSound (soundFile: string);
procedure SyncWaveSound (soundFile: string);
procedure StopWaveSound;
procedure SetWaveSound (switch: boolean);
procedure WaveSoundVolume (volume: word);                            overload;
procedure WaveSoundVolume (volLeft, volRight: word);                 overload;
procedure SetWavePlaybackRate (pitch: float);
procedure ShowWaveError (mmResult: Longword; funcName: string);
procedure ErrorBeep;
procedure WarningBeep;

// time handling
procedure GetTime (var Hour, Min, Sec, MSec: word);
procedure GetDate (var Year, Month, Day: word);
function SystemTimeString: string;
function SystemDateString: string;
function SystemDatumString: string;
// MSec time measuring 1
procedure MarkTime;
function GetRuntimeMS: longWord;
function PictureRuntimeMS: integer;
function PictureRuntime: extended;
//function RunTimeMS: longWord;
function PicturesPerSecond: float;
function FPS: string;
// MSec time measuring 2
procedure MarkTime2;
//function RunTimeMS2: longWord;
function GetRuntimeMS2: longWord;
function FPS2: string;

// Delphi development version
function DelphiIsRunning: boolean;
function DelphiVersion: string;

// Operation System information
function OperatingSystem: string;
function WindowsVersion: string;
function WindowsProductID: string;
function GetComputername: string;
function GetProgramName: string;
function CurrentUserName: string;
function OSLanguage: string;
function DirectXVersion: string;
function DefaultPrinterName: string;

// Hardware information
function CPUSpeed: integer;   // works only for Pentium and higher
function CPUSpeed2: integer;
function ProcessorType: string;
function ProcessorName: string;
procedure GetHardwareInfo (out hardwareInfo: tHardwareInfo);

// not useable if Win32PlatForm = VER_PLATFORM_WIN32_NT
procedure SetPort (portAddr: word; value: byte); assembler; stdcall;
function  GetPort (portAddr: word): byte; assembler; stdcall;

// memory handling
procedure FreeAndNil (var Obj);
function MemoryStatus: TMemoryStatus;
function TotalPhysicalMemory: Int64;
function AvailablePhysicalMemory: Int64;
function TotalVirtualMemory: Int64;
function AvailableVirtualMemory: Int64;

// screen information
procedure GetDisplayInfo (out screenInfo: tDeviceInfo);
function DisplayInfoString: string;
function DisplayInfoString2: string;
procedure ResetDisplayMode;
function SetDisplayMode (index: integer): longint;
procedure ChangeDisplayMode (index: integer);
function GetDisplayModeList (colorBits: byte = 0): tStringList;

// window handling
procedure ShowTaskBar (show: boolean);

// drive handling
function GetDriveList: TStringList;
function DriveReady (const drive: byte; var size,free: Int64): Boolean;
function GetVolumeName (drive: string): string;
function GetCDList: shortString;

// file handling
procedure RestoreStartDirectory;
procedure GetFileList (var sList: TStringList; const path, ext: string);
function GetFileSize     (const filename: String): integer;
function FileSizeByName  (const filename: string): integer;
function IsFileWriteable (const filename: string): boolean;
function SubstituteFilename (filename: string): string;
function GetFilenameExtension (const filename: string): string;
function RemoveFilenameExt    (const fileName: string): string;
function CopyFile (const existingFileName: string;
                   const newFileName: string;
                   failIfExists: boolean): boolean;

// program execution
function InitMutex (name: string): boolean;
procedure ClearMutex;
procedure StopMultiStartedProgram;
procedure DOSCommand (command: string);
function RunApplication (const name, param: string): boolean;
procedure RunControlPanelApplet (appletFileName: string);
procedure RunLink (link: string);

// display messages, warnings and errors
procedure RaiseObjectNilError  (name: string);
procedure RaiseObjectInitError (name: string);
procedure RaiseMemoryError     (name: string);
procedure RaiseProcedureError  (name: string);

procedure ShowMessageBox (const caption, infoText: string);
procedure ShowHint       (const caption, infoText: string);
procedure ShowWarning    (const caption, infoText: string);
function  ShowError      (const caption, infoText: string;
                          const buttons: word = $102): integer;
function  ShowLastError  (const caption, infoText: string;
                          const buttons: word = $102): integer;
procedure ShowFatalError (const exitCode: byte;
                          const caption, infoText: string);
function ShowFileError (hide: byte; filename: string): boolean;


{=== stack routines ===}

function PushInteger (const value: integer): integer;
function PopInteger: integer;
function PushBoolean (const value: boolean): boolean;
function PopBoolean: boolean;
function PushFloat (const value: float): float;
function PopFloat: float;

(*  function Pop can not be overloaded !!!
function Push(const value: integer): integer;                        overload;
function Pop: integer;                                               overload;
function Push(const value: boolean): boolean;                        overload;
function Pop: boolean;                                               overload;
function Push(const value: float): float;                            overload;
function Pop: float;                                                 overload;
*)

{=== TObjectStack class ===}

//---------------------------------------------------------
// complete replacement of class TObjectStack of unit Contnrs
// is faster because of compact coding of the class.
//---------------------------------------------------------
// kompletter Ersatz der Klasse TObjectStack der Unit Contnrs
// ist etwas schneller wegen der kompakten Kodierung der Klasse.
//---------------------------------------------------------

type  TObjectStack = class(TObject)
      public
        constructor Create;
        destructor Destroy; override;
        procedure Clear;
        function Count: integer;
        function Push (aObject: TObject): TObject;
        function Pop: TObject;
        function Peek: TObject;
        function AtLeast (aCount: integer): boolean;
      protected
        FList: TList;
        procedure PushItem(aItem: pointer); virtual;
        function PopItem: pointer; virtual;
        function PeekItem: pointer; virtual;
//        property List: TList read FList;
      public
//        property Count: integer read FList.Count;
      end;

//=== global values ===

var   programName: string;
      computerName: string[MAX_COMPUTERNAME_LENGTH];
      ioError: integer = 0;            // last file error
      memStat: TMemoryStatus;          // memory status
      msgBoxIsOpen: boolean = false;   // true, if modal message box is displayed
      picCnt: longWord = 0;            // picture counter:
      startDirectory: string = '';     // actual directory at program start
      startDisplayMode: integer = 0;   // display mode at program start
      screenInfo:      tDeviceInfo;    // last screen device information
      startScreenInfo: tDeviceInfo;    // screen information at program start
                                       // used for FPS calculation
                                       // reseted on key input and step mode

      mmSoundCard: boolean;            // Multimedia soundcard yes/no
      mmSoundOn: boolean;              // Multimedia sound on/off
      mmSoundFile: string;             // Multimedia sound file
      mmResult: longword;              // Multimedia call result
      mmWaveOutDevices: integer;       // Multimedia waveOut devices
      mmWaveOutCaps1: WAVEOUTCAPS;     // Multimedia waveOut device 1 capabilities
      mmWaveOutCaps2: WAVEOUTCAPS;     // Multimedia waveOut device 2 capabilities

{--- debug file, if compiler option is set to D+ ---}
const debugFilename = 'DEBUG.TMP';
{$ifopt D+}
  var   tst: text;
{$endif}

//*********************************************************
implementation

uses  Clipbrd, Messages, Dialogs, Registry, ShellApi;

//*********************************************************
// return random number between given min and max value
//---------------------------------------------------------
function random2 (min,max: integer): integer; overload;
begin
  result := min + system.random (max - min);
end;
//---------------------------------------------------------
function random2 (min,max: float): float; overload;
begin
  result := min + system.random * (max - min);
end;

//=== get minima or maxima ===
//=== ermittle Minimum oder Maximum ===

//*********************************************************
// get minimum value   /   Minimalwert ermitteln
//---------------------------------------------------------
function Min (const a,b: longWord): longWord;
begin
  if a < b   then result := a   else result := b;
end;
//---------------------------------------------------------
{$ifdef useAsm}
function Min (const a,b: longInt): longInt;   assembler;
asm
  cmp   edx,eax     // compare (b,a)
  cmovl eax,edx     // b<a -> a:=b
end;
(*
asm                          // old version
  cmp  eax, edx
  jle @return                // a < b -> result = a
  mov  eax, edx              // result = b
 @return:
end;
*)
{$else}
//---------------------------------------------------------
function Min (const a,b: longInt): longInt;
begin
  if a < b   then result := a   else result := b;
end;
{$endif}
{---------------------------------------------------------}
function Min (const a,b: float): float;
begin
  if a < b   then result := a   else result := b;
end;
//*********************************************************
// get minimum of 3 values   /   Minimalwert 3er Werte ermitteln
//---------------------------------------------------------
function Min (const a,b,c: integer): integer;
asm
  cmp   edx,eax     // compare (b,a)
  cmovl eax,edx     // a>b -> a:=b
  cmp   ecx,eax     // compare (c,a)
  cmovl eax,ecx     // a>c -> a:=c
end;

//*********************************************************
// get maximum value   /   Maximalwert ermitteln
//---------------------------------------------------------
function Max (const a,b: longWord): longWord;
begin
  if a > b   then result := a   else result := b;
end;
//---------------------------------------------------------
{$ifdef useAsm}
function Max (const a,b: longInt): longInt;   assembler;
asm
  cmp   edx,eax     // compare (b,a)
  cmovg eax,edx     // b>a -> a:=b
end;
(*
asm
  cmp  eax, edx
  jge @return
  mov  eax, edx
 @return:
end;
*)
{$else}
//---------------------------------------------------------
function Max (const a,b: longInt): longInt;
begin
  if a > b   then result := a   else result := b;
end;
{$endif}
{---------------------------------------------------------}
function Max (const a,b: float): float;
begin
  if a > b   then result := a   else result := b;
end;
//*********************************************************
// get maximum of 3 values   /   Maximalwert 3er Werte ermitteln
//---------------------------------------------------------
function Max (const a,b,c: integer): integer;
asm
  cmp   edx,eax     // compare (b,a)
  cmovg eax,edx     // b>a -> a:=b
  cmp   ecx,eax     // compare (c,a)
  cmovg eax,ecx     // c>a -> a:=c
end;


//=== return value between min and max value ===
//=== Wert zwischen min und max zurckgeben ===

//*********************************************************
// Clamp: return value between min and max
//        (see also Math.EnsureRange)
// note:  value will be only returned as result value !!!
//---------------------------------------------------------
{$ifdef useAsm}
//---------------------------------------------------------
function Clamp (const value, min, max: byte): byte; assembler;
asm                          // AL = value;  DL = min;  CL = max
  cmp  al, dl
  jge @greater               // value > min  -> goto greater
  mov  al, dl                // result = min
  jp  @return
 @greater:
  cmp  al, cl
  jle @return                // value <= max -> return
  mov  al, cl                // result = max
 @return:
end;
//---------------------------------------------------------
function Clamp (const value, min, max: longint): longint; assembler;
asm                          // EAX = value;  EDX = min;  ECX = max
  cmp  eax, ecx
  jle @smaller               // value <= max -> jump
  mov  eax, ecx              // result = max
@smaller:
  cmp  eax, edx
  jge @return                // value > min  -> jump
  mov  eax, edx              // result = min
@return:
end;
{$else}
//---------------------------------------------------------
function Clamp (const value, min, max: byte): byte;
begin
  if value < min then result := min
  else if value > max then result := max
  else result := value;
end;
//---------------------------------------------------------
function Clamp (const value, min, max: longint): longint;
begin
  if value < min then result := min
  else if value > max then result := max
  else result := value;
end;
{$endif}
//---------------------------------------------------------
function Clamp (const value, min, max: int64): int64;
begin
  if      value < min then result := min
  else if value > max then result := max
  else result := value;
end;
//---------------------------------------------------------
function Clamp (const value, min, max: float): float;
begin
  if      value < min then result := min
  else if value > max then result := max
  else result := value;
end;

//---------------------------------------------------------
// ClampByte: clamp to byte value
//            return value between 0 and 255
//---------------------------------------------------------
{$ifdef useAsm}
(*
function ClampByte (value: longInt): byte; assembler;
asm
  mov  eax, value  // store value in EAX register (32-bit register)
  cmp  eax, 1      // if less than 255, compare to 1
  jl  @setlo       // if less than 1, set it to 0 (min value)
  cmp  eax, 254    // compare it to 254
  jg  @sethi       // if greater than 254 then go set to 255 (max value)
  ret              // otherwise it doesn't change, just exit
@sethi:            // set value to 255
  mov  eax, 255    // move 255 into the EAX register
  ret              // exit (result value is the EAX register value)
@setlo:            // set value to 0
  mov  eax,0       // move 0 into EAX register
end;               // result is in EAX
*)
//---------------------------------------------------------
function ClampByte (value: longInt): byte; assembler;
asm
  mov   edx,255
  cmp   edx,eax     // compare (255,a)
  cmovl eax,edx     // a>255 -> a:=255

  xor   edx,edx     // edx := 0
  cmp   edx,eax     // compare (0,a)
  cmovg eax,edx     // a<0 -> a:=0
end;                // result a is in EAX
//---------------------------------------------------------
function ClampByte (value: float): byte;
begin
  if      value <   0.0 then result := 0
  else if value > 255.0 then result := 255
  else                       result := round (value);
end;
//---------------------------------------------------------
{$else}
function ClampByte (value: longInt): byte;
begin
  if value < 0 then result := 0
  else if value > 255 then result := 255
  else result := value;
end;
{$endif}
//---------------------------------------------------------
// ClampWord: clamp to word value
//            return value between 0 and 65535
//---------------------------------------------------------
function ClampWord (value: longInt): word;
begin
  if value < 0 then result := 0
  else if value > maxWord then result := maxWord
  else result := value;
end;

//*********************************************************
// swap values   /   Werte vertauschen
//---------------------------------------------------------
{$ifdef useAsm}
procedure Swap (var a,b: byte);   assembler;
asm
  mov  cl, byte Ptr [EAX]
  xchg cl, byte Ptr [EDX]
  mov  byte Ptr [EAX], cl
end;
//---------------------------------------------------------
procedure Swap (var a,b: shortInt);   assembler;
asm
  mov  cl, byte Ptr [EAX]
  xchg cl, byte Ptr [EDX]
  mov  byte Ptr [EAX], cl
end;
//---------------------------------------------------------
procedure Swap (var a,b: word);   assembler;
asm
  mov  cx, word Ptr [EAX]
  xchg cx, word Ptr [EDX]
  mov  word Ptr [EAX], cx
end;
//---------------------------------------------------------
procedure Swap (var a,b: smallInt);   assembler;
asm
  mov  cx, word Ptr [EAX]
  xchg cx, word Ptr [EDX]
  mov  word Ptr [EAX], cx
end;
//---------------------------------------------------------
procedure Swap (var a,b: longInt);   assembler;
asm
  mov  ecx, [EAX];   xchg ecx, [EDX];   mov  [EAX], ecx
end;
//---------------------------------------------------------
procedure Swap (var a,b: cardinal);   assembler;
asm
  mov  ecx, [EAX];   xchg ecx, [EDX];   mov  [EAX], ecx
end;
//---------------------------------------------------------
procedure Swap (var a,b: pointer);   assembler;
asm
  mov  ecx, [EAX];   xchg ecx, [EDX];   mov  [EAX], ecx
end;

{$else}
//---------------------------------------------------------
procedure Swap (var a,b: byte);
var   c: byte;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: shortInt);
var   c: shortInt;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: word);
var   c: word;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: smallInt);
var   c: smallInt;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: longInt);
var   c: longInt;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: cardinal);
var   c: cardinal;
begin
  c := a;   a := b;   b := c;
end;
//---------------------------------------------------------
procedure Swap (var a,b: pointer);
var   c: pointer;
begin
  c := a;   a := b;   b := c;
end;
{$endif}
//---------------------------------------------------------
procedure Swap (var a,b: int64);
var   c: int64;
begin
  c := a;   a := b;   b := c;
end;


//===  Sort: return values in ascendent order (a <= b) ===
//===  Sort: Werte aufsteigend zurckgeben (a <= b) ===

//---------------------------------------------------------
procedure Sort (var a,b: smallint);    // -32768..32767
begin
  if a > b then Swap (a,b);
end;
//---------------------------------------------------------
procedure Sort (var a,b: int64);       // -2^63..2^63 -1
begin
  if a > b then Swap (a,b);
end;
{$ifdef useAsm}
//---------------------------------------------------------
procedure Sort (var a,b: byte);   assembler;
asm
  mov  cl, byte Ptr [EAX]
  cmp  cl, byte Ptr [EDX]
  jle @return
  xchg cl, byte Ptr [EDX]
  mov  byte Ptr [EAX], cl
 @return:
end;
//---------------------------------------------------------
procedure Sort (var a,b: word);   assembler;
asm
  mov  cx, word Ptr [EAX]
  cmp  cx, word Ptr [EDX]
  jle @return
  xchg cx, word Ptr [EDX]
  mov  word Ptr [EAX], cx
 @return:
end;
//---------------------------------------------------------
procedure Sort (var a,b: longInt);   assembler;
asm                          // EAX = a;  EDX = b
  mov  ecx, [EAX]            // ECX = a
  cmp  ecx, [EDX]            // a <> b
  jle @return                // a <= b ->  exit
  xchg ecx, [EDX]            // a <-> b
  mov  [EAX], ecx            // a = ECX
 @return:
end;
//---------------------------------------------------------
procedure Sort (var a,b: longWord);   assembler;
asm                          // EAX = a;  EDX = b
  mov  ecx, [EAX]            // ECX = a
  cmp  ecx, [EDX]            // a <> b
  jle @return                // a <= b ->  exit
  xchg ecx, [EDX]            // a <-> b
  mov  [EAX], ecx            // a = ECX
  @return:
end;
{$else}
//---------------------------------------------------------
procedure Sort (var a,b: byte);
begin
  if a > b then Swap (a,b);
end;
//---------------------------------------------------------
procedure Sort (var a,b: word);
begin
  if a > b then Swap (a,b);
end;
//---------------------------------------------------------
procedure Sort (var a,b: longWord);
begin
  if a > b then Swap (a,b);
end;
//---------------------------------------------------------
procedure Sort (var a,b: longInt);
begin
  if a > b then Swap (a,b);
end;
{$endif}


//=== rotate index ===

//*********************************************************
// increase enumeration index
// and if index is over the maximum index set it to minimum index
//---------------------------------------------------------
procedure IncIndex (var index: integer; const min,max: integer; delta: integer = 1);
begin
  index := index + 1;
  if index > max then index := min;
end;
//*********************************************************
// decrease enumeration index
// and if index is under the minimum limit set it to maximum index
//---------------------------------------------------------
procedure DecIndex (var index: integer; const min,max: integer; delta: integer = -1);
begin
  index := index + delta;
  if index < min then index := max;
end;

//=== arithmetic shift operations ===

function SAR_2 (value: integer): integer;   asm   SAR EAX,2    end;
function SAR_4 (value: integer): integer;   asm   SAR EAX,4    end;
function SAR_8 (value: integer): integer;   asm   SAR EAX,8    end;
function SAR_9 (value: integer): integer;   asm   SAR EAX,9    end;
function SAR_12(value: integer): integer;   asm   SAR EAX,12   end;
function SAR_13(value: integer): integer;   asm   SAR EAX,13   end;
function SAR_14(value: integer): integer;   asm   SAR EAX,14   end;
function SAR_15(value: integer): integer;   asm   SAR EAX,15   end;
function SAR_16(value: integer): integer;   asm   SAR EAX,16   end;

//=== pointer arithmetic ===

//*********************************************************
// Zeigerwert um byteSize erhhen und zurckgeben
//---------------------------------------------------------
function PointerAdd (ptr: pointer; byteSize: integer): pointer;
begin
  result := pointer (longint (ptr) + byteSize);
end;
//*********************************************************
// return pointer value difference in bytes
// diff = ptr2 - ptr1
//---------------------------------------------------------
function PointerDiff (ptr1, ptr2: pointer): integer;
begin
  PointerDiff := longint(ptr2) - longint(ptr1);
end;


//=== memory ===

//---------------------------------------------------------
// Fast routine to set variable memory to zero value
// input:      dest     destination memory
//             sizeof   destination size in bytes
// example:    ZeroMem (v1, SizeOf(v1Type));
//---------------------------------------------------------
procedure ZeroMem (var dest; sizeof: integer); assembler; register;
asm
  { eax=dest; edx=sizeof }
  push edi       { protect edi }
  mov  edi, eax  { edi=@dest }
  xor  eax, eax  { eax=0 }
  { calc. no. of dwords to store and store them }
  mov  ecx, edx
  shr  ecx, 2
  rep  stosd
  { calc. no. of missing words and store them (actually this is bit 1) }
  mov  ecx, edx
  bt   ecx, 1
  jnc  @stobyte
  stosw

  @stobyte:   {calc. no. of missing bytes and store them (bit 0)}
  bt  ecx, 0
  jnc @ende
  stosb       {zero memory block}
  @ende:
  pop edi
end;
//---------------------------------------------------------
// Fast swap routine that swaps the contents of any 2 variables.
// The variables may be of any type but the size
// of the variables must be passed in length
// eg.  v1,v2: integer;    SwapMem (v1,v2, SizeOf(integer));
//---------------------------------------------------------
procedure SwapMem (var source, dest; len: integer);
asm               //   eax     edx   ecx
  push edi
  push esi
  mov esi, source     // esi/eax = source
  mov edi, dest       // edi/edx = dest
  mov ecx, len        // ecx = count
  cld
 @L1:
  mov al,[edi]
  xchg [esi],al
  inc si
  stosb
  loop @L1
  pop esi
  pop edi
end;

//*********************************************************
// fill longword memory with longword value (like FillChar)
// input:      dest     memory destination
//             count    number of fill values
//             value    longword fill value
//---------------------------------------------------------
procedure FillLongWord (var dest; count: integer; value: longword);
asm                     //  eax   edx             ecx
  PUSH    EDI
  MOV     EDI,EAX       // edi/eax = @dest
  MOV     EAX,ECX       // eax = value
  MOV     ECX,EDX       // ecx = count
  TEST    ECX,ECX       // count = 0 ?
  JS      @exit         // count = 0 -> exit
  REP     STOSD         // fill count dwords with value
@exit:
  POP     EDI
end;
//*********************************************************
// copy longword value memory (like Move)
// input:      source   memory soure
//             dest     memory destination
//             count    number of fill values
//---------------------------------------------------------
procedure MoveLongword (const source; var dest; count: Integer);
asm                    //     eax         edx   ecx
  PUSH    ESI
  PUSH    EDI
  MOV     ESI,EAX      // esi/eax = source
  MOV     EDI,EDX      // edi/edx = dest
  MOV     EAX,ECX      // eax = count
  CMP     EDI,ESI      // source = dest ?
  JE      @exit        // source = dest -> exit
  REP     MOVSD        // copy count values from source to destination
@exit:
  POP     EDI
  POP     ESI
end;


//=== string handling ===

//*********************************************************
// search 'substr' in 's' from right to left and return
// string position (=0 if 'substr' not found)
//---------------------------------------------------------
// Sucht 'substr' in 's' von rechts nach links und gibt
// die Position zurck (=0 wenn 'substr' nicht gefunden).
//---------------------------------------------------------
//   sp := RightPos ('123.456.789', '.');
//   -->  sp = 8
//---------------------------------------------------------
function RightPos (s: string; substr: string): byte;

var   si: byte;
begin
  RightPos := 0;
  for si := length(s) downto 1
  do if substr = s[si]
  then begin
    RightPos := si;
    break;
  end;
end;
//*********************************************************
// search 'substr' in 's' from left to right and return
// string left of 'substr'. if substr not exist return whole string
//---------------------------------------------------------
// Sucht 'substr' in 's' von links nach rechts und gibt
// den String links von 'substr' zurck.
//---------------------------------------------------------
//  s := GetLeftString ('filename.exe abs', '.exe');  // s = 'filename'
//  s := GetLeftString ('nosymbol', '.');             // s = 'nosymbol'
//---------------------------------------------------------
function GetLeftString (s: string; substr: string): string;
var   si: byte;
begin
  si := Pos (substr, s);
  if si > 0
  then Result := Copy (s, 1, si-1)
  else Result := s;
end;
//*********************************************************
// search 'substr' in 's' from right to left and return
// string right of 'substr'.
//---------------------------------------------------------
// Sucht 'substr' in 's' von rechts nach links und gibt
// den String rechts von 'substr' zurck.
//---------------------------------------------------------
//   s := GetRightString ('C:\path\filename.ext', '\');
//   -->  s = 'filename.ext'
//---------------------------------------------------------
function GetRightString (s: string; substr: string): string;
var   si: byte;
begin
  si := RightPos (s, substr);
  if si > 0
  then Result := Copy (s, si+1, 255)
  else Result := '';
end;

//*********************************************************
// return string without blancs
// see also SysUtils.Trim, TrimLeft, TrimRight
// input:      s1       string to trim
// result:     string   trimmed string
//---------------------------------------------------------
function TrimAll (const s: string): string;
var   sp: integer;
begin
  result := s;
  sp := pos (' ', result);
  while sp > 0 do begin
    delete (result, sp, 1);
    sp := pos (' ', result);
  end;
end;
//*********************************************************
// return string without given substrings
// input:      s        string to trim
//             subStr   delete all this substrings
// result:     string   trimmed string
// example:    s := TrimAll(s, ' ');   // delete all blancs of s 
//---------------------------------------------------------
function TrimAll (const s, subStr: string): string;
var   sp: integer;
begin
  result := s;
  sp := pos (substr, result);
  while sp > 0 do begin
    delete (result, sp, length(substr));
    sp := pos (substr, result);
  end;
end;
//*********************************************************
// check if string is a valid longint value
// input:      s        string to check
// result:     boolean  =true if string is a longint value
// see also:   SysUtils.StrToInt
// example:    -12345678 = true    +12345678901 = false
//---------------------------------------------------------
var  iVal, errorPos: Integer;
//---------------------------------------------------------
function IsInteger (s: string): Boolean;
//var  iVal, errorPos: Integer;   //Warnings
begin
  Val (s, iVal, errorPos);
  Result := (errorPos = 0);
end;
//*********************************************************
// check if string is a valid int64 value
//---------------------------------------------------------
function IsInt64 (s: string): Boolean;
//var  iVal, errorPos: Integer;   //Warnings
begin
  Val (s, iVal, errorPos);
  Result := (errorPos = 0);
end;

//*********************************************************
// return true if string is a valid float value
// input:      s        string to check
// result:     boolean  =true if string is a float value
// example:    -1,23E-22 = true     1.23E+22
//---------------------------------------------------------
function IsFloat (s: string): boolean;
var  f: extended;
begin
  if pos('.', s) > 0 then s[pos('.', s)] := ',';
  Result := SysUtils.TextToFloat (PChar(s), f, fvExtended);
end;
//*********************************************************
(*
function IsFloat (s: string): boolean;
var  fVal: float;   errorPos: integer;
begin
  if pos(',', s) > 0 then s[pos(',', s)] := '.';
  Val (s, fVal, errorPos);
  Result := (errorPos = 0);
end;
*)

//=== convertions / Konvertierungen ===

//*********************************************************
// return value as binary string
//---------------------------------------------------------
// gibt den Wert als binren Text (0|1) zurck
//---------------------------------------------------------
function BinString (value: byte): string;
var   si: word;
begin
  result := '';
  for si := 7 downto 0
  do result := result + char (((value shr si) and $1) + 48);
end;
//---------------------------------------------------------
function BinString (value: word): string;
begin
  result := BinString (byte(value shr 8)) + '.' + BinString (byte(value));
end;
//---------------------------------------------------------
function BinString (value: longint): string;
begin
  result := BinString (word (value shr 16)) + '.' + BinString (word (value));
end;
//---------------------------------------------------------
function BinString (value: int64): string;
begin
  result := BinString (longint(value shr 32)) + ':' + BinString (longint(value));
end;

//*********************************************************
// return value as hexadecimal string
// see also SysUtils.IntToHex
//---------------------------------------------------------
// gibt den Wert als Hexadezimalen Text zurck
//---------------------------------------------------------
function HexString (value: byte): string;
begin
  result := hexChar[value shr 4] + hexChar[value and 15];
end;
//---------------------------------------------------------
function HexString (value: word): string;
begin
  result := HexString (byte(value shr 8)) + HexString (byte(value));
end;
//---------------------------------------------------------
function HexString (value: longWord): string;
begin
  result := HexString (word(value shr 16)) + HexString (word(value));
end;
//---------------------------------------------------------
function HexString (value: longWord; sep: string): string;
begin
  result := HexString (word(value shr 16)) + sep + HexString (word(value));
end;
//---------------------------------------------------------
function HexString (value: longInt): string;
begin
  result := HexString (word(value shr 16)) + HexString (word(value));
end;
//---------------------------------------------------------
function HexString (value: longInt; sep: string): string;
begin
  result := HexString (word(value shr 16)) + sep + HexString (word(value));
end;
//---------------------------------------------------------
function HexString (value: int64): string;
begin
  result := HexString (longWord(value shr 32)) + HexString (longWord(value));
end;
//---------------------------------------------------------
function HexString (value: int64; sep: string): string;
begin
  result := HexString (longWord(value shr 32)) + sep + HexString (longWord(value));
end;
//---------------------------------------------------------
function HexString (value: pointer): string;
begin
  result := HexString (longWord(value), ':');
end;
//---------------------------------------------------------
function HexString (value: tProcedure): string;
begin
  result := HexString (longWord(@value));
end;
//*********************************************************
// convert hexadecimal string to int64 long integer value
//---------------------------------------------------------
// konvertiert Hexadezimalen Text in eine int64 long integer Wert
// siehe auch SysUtils.HexToBin
//---------------------------------------------------------
function HexToLong (text: string): int64;
var   i1: word;
      hv: int64;
begin
  hv := 0;
  if text[1] = '$' then delete (text,1,1);
  try
    For i1 := 1 to length (text)
    do hv := hv shl 4 + pos (upCase (text[i1]), hexChar) - 1;
  except
  end;
  result := hv;
end;

//*********************************************************
// convert integer value to string
// see also:   SysUtils.IntToStr, SysUtils.StrToIntDef
// input:      value     integer value
//             digits    optional minimal string length
//                >0: text length
//                =0: length as needed
//                <0: with leading zeros
// result:     string;    integer value string
//---------------------------------------------------------
// Aufruf:     txt := IntString (value, digits);
// siehe auch: SysUtils.IntToStr
// Funktion:   Der angegebene Integer - Wert wird in einen
//             Text mit angegebener Lnge konvertiert.
// Versorgung: value: integer;    zu konvertierender Wert
//             digits: integer;   optionelle Textlnge
//               >0: Lnge des Textes
//               =0: bentigte Lnge
//               <0: mit fhrende Nullen
// Ergebnis:   string;    Integer-Wert als Text
//---------------------------------------------------------
function IntString (value: integer; digits: integer = 0): string;

var   s: string[80];   i: word;   nullen: boolean;
begin
  if digits < 0
  then begin
    digits := abs (digits);
    nullen := true;
  end
  else nullen := false;
  if value = -maxint then s := StringOfChar (' ',digits)
  else begin
    str (value:digits, s);
    if nullen then for i := 1 to length (s)
    do if s[i] = ' ' then s[i] := '0';
  end;
  IntString := s;
end;
//*********************************************************
// convert integer value to string with dot notation
// see also:   SysUtils.IntToStr
// input:      value     integer value
//             digits    optional minimal string length
// result:     string    integer value string
//---------------------------------------------------------
// Konvertiert einen Integer-Wert in einen String
// siehe auch: SysUtils.IntToStr
// Eingabe:    value     zu konvertierender Integer-Wert
//             digits    Textlnge
// Rckgabe:   string    Integer-Wert als Text
//---------------------------------------------------------
function IntDotString (value: integer; digits: integer = 0): string;
var   sp: integer;
begin
  str (value, result);
  sp := 3;
  while length(result) > sp
  do begin
    Insert ('.', result, length(result) - sp + 1);
    inc (sp,4);
  end;
  while length(result) < digits do result := ' ' + result;
end;
//*********************************************************
// convert float value with given precision to string
// see also:   SysUtils.FloatToStr, .FloatToStrF, .FloatToText
// input:      value       float value to convert
//             precision   number of decimals after point
//             digits      string length
// result:     string      float value string
//---------------------------------------------------------
// Konvertiert einen Float-Wert in einen String
// siehe auch: SysUtils.FloatToStr, .FloatToStrF, .FloatToText
// Eingabe:    value       zu konvertierender Gleitkommawert
//             precision   Anzahl der Nachkommastellen
//             digits      Textlnge
// Rckgabe:   string      Float-Wert als Text
//---------------------------------------------------------
function FloatString (value: Extended; precision: integer = 0;
                                       digits: integer = 0): string;
begin
  if (precision = 0) and (digits = 0)
  then result := SysUtils.FloatToStr(value)
  else Str (value:digits:precision, result);
end;

//*********************************************************
// convert float strings with dot notation to value
// example:  1.2E45  = valid float value
//---------------------------------------------------------
function StrToFloat2 (fs: string): Extended;
var   ni, cv: integer;
begin
  for ni := length (fs) downto 1
  do begin
    cv := ord(fs[ni]);
    if (cv < ord('0')) or (cv > ord('9'))
    then if (cv <> ord('.')) and (cv <> ord(','))
    then delete(fs, ni,1);
  end;
  if pos ('.', fs) > 0 then fs[pos('.',fs)] := ',';
  if fs = ''
  then result := 0.0
  else result := SysUtils.StrToFloat(fs);
end;
//*********************************************************
// convert boolean value to 'true' or 'false' string
//---------------------------------------------------------
function TrueFalse (flag: boolean): string;
begin
  if flag then result := 'true ' else result := 'false';
end;
//*********************************************************
// convert boolean value to 'on' or 'off' string
//---------------------------------------------------------
function OnOff (flag: boolean): string;
begin
  if flag then result := 'on' else result := 'off';
end;
//*********************************************************
// convert boolean value to 'yes' or 'no' string
//---------------------------------------------------------
function YesNo (flag: boolean): string;
begin
  if flag then result := 'yes' else result := 'no';
end;


//=== sound handling ===

//*********************************************************
// play sound with given frequency on interal beeper
// see also:   Windows.Beep (hz, time);
//             Windows.MessageBeep (type);
//---------------------------------------------------------
procedure Sound (Hz: word);
begin
  if Win32PlatForm = VER_PLATFORM_WIN32_NT
  then begin
    windows.Beep (Hz, 10);
  end
  else begin
    if Hz < 19 then Hz := 19;
    SetPort ($61, GetPort ($61) or 3);      // enable speaker
    hz := word (1193181 div longInt(hz));   // 18.206... * 65536 = 1193180
    SetPort ($43, $B6);      // set timer 2 to square wave generator
    SetPort ($42, Lo(hz));   // set frequence
    SetPort ($42, Hi(hz));
  end;
end;
//*********************************************************
// turn off sound
//---------------------------------------------------------
procedure NoSound;
begin
  if Win32PlatForm = VER_PLATFORM_WIN32_NT
  then windows.Beep (0,0)
  else SetPort ($61, GetPort($61) and $FC);
end;


//=== multimedia sound handling ===

//*********************************************************
// initialize multimedia sound parameter
//---------------------------------------------------------
procedure InitSound;
begin
  MMSystem.TimeBeginPeriod (0);   // set minimum timer resolution in ms
  mmWaveOutDevices := MMSystem.WaveOutGetNumDevs;
  mmSoundCard := mmWaveOutDevices > 0;
  if mmWaveOutDevices > 0
  then mmResult := WaveOutGetDevCaps (0, @mmWaveOutCaps1, sizeof(mmWaveOutCaps1));
  if mmWaveOutDevices > 1
  then mmResult := WaveOutGetDevCaps (1, @mmWaveOutCaps2, sizeof(mmWaveOutCaps2));
  ShowWaveError (mmResult, 'WaveOutGetDevCaps');
end;
//*********************************************************
// play error sound
//---------------------------------------------------------
procedure ErrorBeep;
begin
  MessageBeep (MB_ICONERROR);
end;
//*********************************************************
// play warning sound
//---------------------------------------------------------
procedure WarningBeep;
begin
  MessageBeep (MB_ICONWARNING);
end;
//*********************************************************
// store name of sound file
//---------------------------------------------------------
procedure NameWaveSound (soundFile: string);
begin
  mmSoundFile := soundFile;
end;
//*********************************************************
// play wave sound given by the file name at the background
// and loop until StopWaveSound is called
// example:    PlayWaveSound ('Utopia Windows starten.WAV');
//---------------------------------------------------------
procedure LoopWaveSound (soundFile: string);
begin
  if MMSystem.sndPlaySound (pChar(soundFile),
                       SND_ASYNC or SND_LOOP or SND_FILENAME)
  then mmSoundOn := true;
  mmSoundFile := soundFile;
end;
//*********************************************************
// play wave sound given by the file in the background
// input:   soundFile     filename of wave file
//---------------------------------------------------------
procedure PlayWaveSound (soundFile: string);
begin
  MMSystem.sndPlaySound (pChar(soundFile), SND_ASYNC); // or SND_FILENAME);
end;
//*********************************************************
// play wave sound given by the file name and wait ...
// input:   soundFile     filename of wave file
//---------------------------------------------------------
procedure SyncWaveSound (soundFile: string);
begin
  MMSystem.sndPlaySound (pChar(soundFile), SND_SYNC or SND_FILENAME);
end;
//*********************************************************
// stop playing background wave sound
//---------------------------------------------------------
procedure StopWaveSound;
begin
  MMSystem.sndPlaySound (nil, 0);
  mmSoundOn := false;
end;
//*********************************************************
// turn last wave sound on/off
//---------------------------------------------------------
procedure SetWaveSound (switch: boolean);
begin
  mmSoundOn := switch;
  if mmSoundOn
  then LoopWaveSound (mmSoundFile)
  else StopWaveSound;
end;
//*********************************************************
// set wave sound volume to 0..100 %
//---------------------------------------------------------
procedure WaveSoundVolume (volume: word);
begin
  mmResult := MMSystem.WaveOutSetVolume (integer(wave_mapper),
                         (volume * 655) shl 16 + volume * 655);
  ShowWaveError (mmResult, 'WaveOutSetVolume');
end;
//*********************************************************
// set wave sound volume of left and right channel
// to 0..100 %
//---------------------------------------------------------
procedure WaveSoundVolume (volLeft, volRight: word);
begin
  mmResult := MMSystem.WaveOutSetVolume (integer(wave_mapper),
                         (volRight * 655) shl 16 + volLeft * 655);
  ShowWaveError (mmResult, 'WaveOutSetVolume');
end;
//*********************************************************
// set wave out playback rate: 0x00010000 = 1.0
//                             0x00008000 = 0.5
// http://wwwswt.informatik.uni-rostock.de/deutsch/Mitarbeiter/michael/lehre/Delphi_WS2000/Vortrag06/#MMSYSTEM
//---------------------------------------------------------
procedure SetWavePlaybackRate (pitch: float);
//var   hWaveOut: integer;
var   WaveFormat: TPCMWaveFormat;
begin
//  hWaveOut := 0;
  ZeroMem (WaveFormat, sizeof(WaveFormat));
//  mmResult := waveOutOpen (@hWaveOut, 0, @WaveFormat, 0,0, wave_mapped);
//  mmResult := MMSystem.waveOutSetPlaybackRate (hWaveOut, round(pitch * $10000));
  ShowWaveError (mmResult, 'waveOutOpen');
end;
//*********************************************************
// display last wave out error
//---------------------------------------------------------
procedure ShowWaveError (mmResult: Longword; funcName: string);
var   errorText: string[255];
begin
  if mmResult = 0 then exit;
  errorText[0] := #255;
  MMSystem.waveOutGetErrorText(mmResult, @errorText[1], 255);
  ShowError ('SysUtil2: MMSystem.' + funcName,  errorText);
end;

//=== time handling ===

//*********************************************************
// return actual system time values
//---------------------------------------------------------
procedure GetTime (var Hour, Min, Sec, MSec: word);
var   systemTime: TSystemTime;
begin
  Windows.GetLocalTime (systemTime);
  hour := systemTime.wHour;
  min  := systemTime.wMinute;
  sec  := systemTime.wSecond;
  mSec := systemTime.wMilliSeconds;
end;
//*********************************************************
// return actual system date values
//---------------------------------------------------------
procedure GetDate (var Year, Month, Day: word);
var   systemTime: TSystemTime;
begin
  Windows.GetLocalTime(SystemTime);
  Year  := systemTime.wYear;
  Month := systemTime.wMonth;
  Day   := systemTime.wDay;
end;

//*********************************************************
// return actual system time as string [hh:mm:ss]
// example:    timeStr := SystemTimeString;
//---------------------------------------------------------
function SystemTimeString: string;
var   SystemTime: TSystemTime;
begin
  Windows.GetLocalTime(SystemTime);
  with systemTime
  do result := Format('%d:%2.2d:%-2.2d',[wHour,wMinute,wSecond]);
end;
//*********************************************************
// return actual system date as string [mm-dd-yyyy]
// example:    dateStr := SystemDateString;
//---------------------------------------------------------
function SystemDateString: string;
var   SystemTime: TSystemTime;
begin
  Windows.GetLocalTime(SystemTime);
  with systemTime
  do result := Format('%2d-%2.2d-%4.4d',[wMonth,wDay,wYear]);
end;
//*********************************************************
// return actual system date as string [dd.mm.yyyy]
// example:    dateStr := SystemDateString;
//---------------------------------------------------------
function SystemDatumString: string;
var   SystemTime: TSystemTime;
begin
  Windows.GetLocalTime(SystemTime);
  with systemTime
  do result := Format('%2d.%2.2d.%4.4d',[wDay,wMonth,wYear]);
end;

//=== MSec time measuring 1 ===

var   t1,t2: longWord;        // time values 0..4294967295

//*********************************************************
// reset picture counter and mark actual time
// t2 holds the number of milliseconds since Windows was started.
// note: elapsed time will wrap around to zero if Windows is run
//       continuously for 49.71 days.
//---------------------------------------------------------
procedure MarkTime;
begin
  picCnt := 0;
  t1 := t2;
//t2 := MMSystem.TimeGetTime;
  t2 := Windows.GetTickCount;
end;
//*********************************************************
// return runtime between last 2 MarkTime calls in milliseconds
//---------------------------------------------------------
function RunTimeMS: longWord;
begin
  result := t2 - t1;
  if result <= 0
  then result := 1;
end;
//*********************************************************
// return runtime since last 'MarkTime' call in milliseconds
//---------------------------------------------------------
function GetRuntimeMS: longWord;
begin
//result := MMSystem.TimeGetTime - t2;
  result := Windows.GetTickCount - t2;
  if result <= 0
  then result := 1;
end;
//*********************************************************
// return average runtime of displaying one picture in milliseconds
// note: picCnt must be incremented after picture drawing !
// result:    integer     last picture display time [milliseconds]
//---------------------------------------------------------
function PictureRuntimeMS: integer;
begin
  if picCnt <= 0 then result := 20
  else result := round (GetRunTimeMS / picCnt);
end;
//*********************************************************
// return average runtime of displaying one picture in seconds
// note: picCnt must be incremented after picture drawing !
// result:    Extended     last picture display time [seconds]
//---------------------------------------------------------
function PictureRuntime: Extended;
begin
  if picCnt <= 0
  then result := 0.020
  else result := GetRunTimeMS / (1000*picCnt);
end;
//*********************************************************
// return pictures (frames) per second (0.1 .. 500)
// note: picCnt must be incremented after picture drawing
//---------------------------------------------------------
function PicturesPerSecond: float;
begin
  result := 1000 * picCnt / GetRunTimeMS;
  result := Clamp (result, 0.1, 500.0);
end;
//*********************************************************
// return frames per second as String
// note: picCnt must be incremented after picture drawing
//---------------------------------------------------------
function FPS: string;
begin
//  result := IntString (1000 * picCnt div GetRunTimeMS);
  result := FloatString (1000 * picCnt / GetRunTimeMS, 1);
end;

//=== MSec time measuring 2 ===

var   ticksPerSecond: int64;      // clockrate = 1193182 tps
      tt1, tt2: int64;
//---------------------------------------------------------
function InitTicker: boolean;
begin
  QueryPerformanceFrequency (ticksPerSecond);
  MarkTime2;
  InitTicker := ticksPerSecond > 0;
end;
//---------------------------------------------------------
procedure MarkTime2;
begin
  QueryPerformanceCounter (tt2);
  tt1 := tt2;
end;
//*********************************************************
// return runtime between last 2 MarkTime calls in milliseconds
//---------------------------------------------------------
function RunTimeMS2: longWord;
begin
  if tt2 > tt1
  then tt2 := tt2 - tt1
  else tt2 := (maxInt64 - tt1) + tt2 + 1;   // timer overflow
  result := longWord ((1000 * tt2) div ticksPerSecond);
  if result <= 0
  then result := 1;
end;
//*********************************************************
// return runtime since last MarkTime call in milliseconds
//---------------------------------------------------------
function GetRuntimeMS2: longWord;
begin
  QueryPerformanceCounter (tt2);
  result := RunTimeMS2;
end;
//*********************************************************
// return frames per second
// note: picCnt must be incremented after picture drawing
//---------------------------------------------------------
function FPS2: string;
begin
  result := IntToStr (1000 * picCnt div GetRunTimeMS2);
end;


//=== development information ===

//*********************************************************
// return true if Delphi IDE is running
//---------------------------------------------------------
function DelphiIsRunning: boolean;
begin
  Result := FindWindow ('TAppBuilder', nil) > 0;
end;
//*********************************************************
// return delphi version
//---------------------------------------------------------
function DelphiVersion: string;
begin
  result := '?.?';
  {$ifdef VER160}  result := '8.0';  {$endif}
  {$ifdef VER150}  result := '7.0';  {$endif}
  {$ifdef VER140}  result := '6.0';  {$endif}
  {$ifdef VER130}  result := '5.0';  {$endif}
  {$ifdef VER120}  result := '4.0';  {$endif}
  {$ifdef VER100}  result := '3.0';  {$endif}
  {$ifdef VER90}   result := '2.0';  {$endif}
  {$ifdef VER80}   result := '1.0';  {$endif}
end;

//=== OS information ===

//*********************************************************
// return name of operating system
//---------------------------------------------------------
function OperatingSystem: string;
begin
  case Win32Platform of
    0: result := 'Windows 3.1';
    1: // 9x-Reihe
      If Win32MajorVersion = 4
      then case Win32MinorVersion of
         0: result:='Windows 95';
        10: result:='Windows 98';
        90: result:='Windows Me';
      end;
    2: // NT-Reihe
      Case Win32MajorVersion of
        3: IF Win32MinorVersion=51 then result:='Windows NT 3.51';
        4: If Win32MinorVersion= 0 then result:='Windows NT 4';
        5: Case Win32MinorVersion of
             0: result := 'Windows 2000';
             1: result := 'Windows XP';
             2: result := 'Windows .NET Server';
           end;
      end;
    else result := 'unknown OS';
  end;
end;
//*********************************************************
// return version of windows operation system
//---------------------------------------------------------
function WindowsVersion: string;
begin
  result := Win32CSDVersion;           // Servicepack info
  if Win32BuildNumber > 0
  then result := result + ', Build ' + IntToStr(Win32BuildNumber);
end;
//*********************************************************
// Gets a string value from the registry from the given root and sub key.
// Converts integer value to a string
// Returns '' if the sub key or value name are not known.
//---------------------------------------------------------
function GetRegistryString (const rootKey: HKEY;
                            const subKey, name: string): string;
var
  reg: TRegistry;            // registry access object
  valueInfo: TRegDataInfo;   // info about registry value
begin
  Result := '';
  reg := TRegistry.Create;
  try
    reg.RootKey := rootKey;
    if reg.OpenKeyReadOnly (subKey)
    then if reg.GetDataInfo (name, valueInfo)  // check if registry value is string or integer
    then begin
      case valueInfo.RegData of
        // string value: just return it
        rdString, rdExpandString:   Result := reg.ReadString (name);
        // integer value: convert to string
        rdInteger:   Result := IntToStr (reg.ReadInteger(name));
      end;
    end;
  finally
    reg.CloseKey;    // close registry
    reg.Free;
  end;
end;
//*********************************************************
// return the Windows Product ID
// you can get it also from InternetExplorer menu ?/info
// or "Icon/Arbeitsplatz / Eigenschaften"
//---------------------------------------------------------
function WindowsProductID: string;
const
  regKey1 = 'Software\Microsoft\Windows\CurrentVersion';
  regKey2 = 'Software\Microsoft\Windows NT\CurrentVersion';
begin
  Result := GetRegistryString (HKEY_LOCAL_MACHINE, regKey1, 'ProductID');
  if result = ''
  then Result := GetRegistryString (HKEY_LOCAL_MACHINE, regKey2, 'ProductID');
end;

//*********************************************************
// return name of computer
//---------------------------------------------------------
function GetComputerName: string;

var   compName: array[0..MAX_COMPUTERNAME_LENGTH] of char;
var   maxSize: longWord;
begin
  maxSize := MAX_COMPUTERNAME_LENGTH;
  windows.GetComputername (compName, maxSize);
  result := compName;
end;
//*********************************************************
// return own program name
//---------------------------------------------------------
function GetProgramName: String;
begin
  result := GetLeftString (GetRightString (ParamStr(0), '\'), '.exe');
end;
//*********************************************************
// return name of current user
//---------------------------------------------------------
function CurrentUserName: String;
var   ca: array[0..127] of Char;   sz: DWord;
begin
  sz := SizeOf (ca);
  GetUserName (ca, sz);
  result := ca;
end;
//*********************************************************
// return Operating System Language
//---------------------------------------------------------
function OSLanguage: string;
var
  LanguageID: LangID;   Language: array [0..100] of char;
begin
  LanguageID := GetSystemDefaultLangID;
  VerLanguageName (LanguageID, Language, 100);
  Result := string (Language);
end;
//*********************************************************
// return DirectX version as string
//---------------------------------------------------------
function DirectXVersion: string;
var    iv: int64;
begin
  result := '?';
  with TRegistry.Create do begin
    try
      RootKey := HKEY_LOCAL_MACHINE;
      if OpenKeyReadOnly ('\Software\Microsoft\DirectX\')
      then if ValueExists ('InstalledVersion')
      then begin
        ReadBinaryData ('InstalledVersion', iv, 8);
        result := IntToStr (iv shr 24) + '.' + IntToStr (iv and $FF);
        if ValueExists ('RC')
        then result := result + char(ord(ReadString('RC')[1]) + 49);
        if ValueExists ('Version')
        then result := result + '   ' + ReadString ('Version');
      end;
    finally
      Free;
    end;
  end;
end;
//*********************************************************
// return default printer name
//---------------------------------------------------------
function DefaultPrinterName: string;
var
  printerName: array[0..255] of char;
begin
  GetProfileString ('Windows','device','',printerName,255);
  Result := GetLeftString (StrPas (printerName), ',');
end;

//*********************************************************
// return CPU speed [MHz] works only for Pentium and higher
//---------------------------------------------------------
function CPUSpeed: integer;

const  delayTime = 200;      // measure time in ms
var    TimerHi, TimerLo: DWORD;
begin
  Sleep(1);
  try
    asm
      dw 310Fh   // rdtsc
      mov TimerLo, eax
      mov TimerHi, edx
    end;
    Sleep (delayTime);
    asm
      dw 310Fh   // rdtsc
      sub eax, TimerLo
      sbb edx, TimerHi
      mov TimerLo, eax
      mov TimerHi, edx
    end;
    result := 1 + TimerLo div (1000 * DelayTime);
  Finally
  End;
end;
//*********************************************************
// return CPU speed [MHz] (with priority handling)
//---------------------------------------------------------
function CPUSpeed2: integer;
var   PriorityClass, Priority : Integer;
begin
  PriorityClass := GetPriorityClass(GetCurrentProcess);
  Priority := GetThreadPriority(GetCurrentThread);

  SetPriorityClass(GetCurrentProcess, REALTIME_PRIORITY_CLASS);
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_TIME_CRITICAL);

  result := CPUSpeed;

  SetThreadPriority(GetCurrentThread, Priority);
  SetPriorityClass(GetCurrentProcess, PriorityClass);
end;
//*********************************************************
// return type of processor as string
//---------------------------------------------------------
function ProcessorType: string;

const PROCESSOR_INTEL = 0;
      PROCESSOR_MIPS  = 1;
      PROCESSOR_ALPHA = 2;
      PROCESSOR_PPC   = 3;

var   SysInfo: TSystemInfo;
      OSInfo: TOSVersionInfo;
begin
  OSInfo.dwOSVersionInfoSize := sizeof(TOSVersionInfo);
  GetVersionEx(OSInfo);
  GetSystemInfo(SysInfo);
  case SysInfo.wProcessorArchitecture of
    PROCESSOR_INTEL:
      case OSInfo.dwPlatformID of
      	VER_PLATFORM_WIN32S, VER_PLATFORM_WIN32_WINDOWS:
          case SysInfo.dwProcessorType of
            3: result := 'Intel 386';
            4: result := 'Intel 486';
            5: result := 'Intel Pentium';
          end;
        VER_PLATFORM_WIN32_NT:
          case SysInfo.wprocessorLevel of
            3: result := 'Intel 386';
            4: result := 'Intel 486';
            5: result := 'Intel Pentium';
          end;
      end;
    PROCESSOR_MIPS:  result := 'MIPS';
    PROCESSOR_ALPHA: result := 'Alpha';
    PROCESSOR_PPC:   result := 'PowerPC';
  end;
end;
//*********************************************************
// return processor name
//---------------------------------------------------------
function ProcessorName: string;
const regKey1 = '\Hardware\Description\System\CentralProcessor\0';
begin
  Result := GetRegistryString (HKEY_LOCAL_MACHINE, regKey1, 'ProcessorNameString');
  if result = ''
  then Result := GetRegistryString (HKEY_LOCAL_MACHINE, regKey1, 'Identifier');
  Result := TrimLeft (Result);
end;
//*********************************************************
// return hardware information
//---------------------------------------------------------
procedure GetHardwareInfo (out hardwareInfo: tHardwareInfo);
begin
  hardwareInfo.processorType := ProcessorType;
  hardwareInfo.processorName := ProcessorName;
  hardwareInfo.CPUSpeed := CPUSpeed;
  hardwareInfo.mousePresent := boolean(GetSystemMetrics(SM_MOUSEPRESENT));
  hardwareInfo.mouseButtons := Integer(GetSystemMetrics(SM_CMOUSEBUTTONS));

end;

//*********************************************************
// not useable if Win32PlatForm = VER_PLATFORM_WIN32_NT
//*********************************************************
procedure SetPort (portAddr: word; value: byte); assembler;
asm
  mov dx, portAddr
  mov al, value
  out dx, al
end;
//*********************************************************
procedure SetPortW (portAddr: word; value: word); register;
asm
  xchg ax,dx
  out dx,ax
end;
//*********************************************************
function GetPort (portAddr: word): byte; assembler;
asm
  mov dx, portAddr
  in  al, dx
end;
//*********************************************************
function GetPortW (portAddr: word): word; register;
asm
  mov dx,ax
  in  ax,dx
end;

//=== memory handling ===

//*********************************************************
// Frees an object reference and replaces the reference with Nil.
// Note: This procedure is a copy of IdGlobal.FreeAndNil
//---------------------------------------------------------
procedure FreeAndNil (var Obj);
var
  P: TObject;
begin
  P := TObject(Obj);
  TObject(Obj) := nil;
  P.Free;
end;

//*********************************************************
// return memory values    see windows.TMemoryStatus
//---------------------------------------------------------
function MemoryStatus: TMemoryStatus;
begin
  MemStat.dwLength:= SizeOf(MemStat);
  GlobalMemoryStatus (MemStat);
  result := MemStat;
end;

//*********************************************************
// physical memory in bytes / Gre des Arbeitsspeichers in Bytes
//---------------------------------------------------------
function TotalPhysicalMemory: Int64;
begin   result := MemoryStatus.dwTotalPhys;  end;
//*********************************************************
// available physical memory in bytes / Freier Arbeitsspeicher
//---------------------------------------------------------
function AvailablePhysicalMemory: Int64;
begin   result := MemoryStatus.dwAvailPhys;   end;
//*********************************************************
// whole virtual memory in bytes / Gre der Auslagerungsdatei
//---------------------------------------------------------
function TotalVirtualMemory: Int64;
begin   result := MemoryStatus.dwTotalVirtual;  end;
//*********************************************************
// available virtual memory in bytes / Freier Platz der Auslagerungsdatei
//---------------------------------------------------------
function AvailableVirtualMemory: Int64;
begin   result := MemoryStatus.dwAvailVirtual;   end;


//=== screen information ===

//*********************************************************
// init starting display mode information
// globals:   startDisplayMode   display mode at program start
//---------------------------------------------------------
procedure InitDisplayInfo;
var
  dm      : Integer;
  deviceMode : TDeviceMode;
begin
  GetDisplayInfo (screenInfo);         // get current screen device data
  startScreenInfo := screenInfo;
  dm := 0;
  startDisplayMode := -1;
  while EnumDisplaySettings(nil, dm, deviceMode)
  do with deviceMode do begin
    if  (dmPelsWidth = startScreenInfo.x)
    and (dmPelsHeight = startScreenInfo.y)
    and (dmBitsPerPel = startScreenInfo.bitsPerPixel)
    and (dmDisplayFrequency = startScreenInfo.refreshRate)
    then startDisplayMode := dm;
    Inc(dm);
  end;
end;
//*********************************************************
// return information about screen and active window
//---------------------------------------------------------
procedure GetDisplayInfo (out screenInfo: tDeviceInfo);
begin
  ZeroMem (screenInfo, sizeof(screenInfo));
  with screenInfo
  do begin
    dcHandle := GetDC(GetActiveWindow);
    // screen resolution
    x := GetDeviceCaps (dcHandle, HORZRES);
    y := GetDeviceCaps (dcHandle, VERTRES);
    // screen size [mm]
    xSize := GetDeviceCaps (dcHandle, HORZSIZE);
    ySize := GetDeviceCaps (dcHandle, VERTSIZE);
    // bits per pixel
    pPlanes := GetDeviceCaps (dcHandle, PLANES);
    bitsPerPixel := GetDeviceCaps (dcHandle, BITSPIXEL) * pPlanes;
    // get aspect ratio
    xAspect := GetDeviceCaps (dcHandle, ASPECTX);
    yAspect := GetDeviceCaps (dcHandle, ASPECTY);

    colRes := GetDeviceCaps (dcHandle, COLORRES);
    dVersion := GetDeviceCaps (dcHandle, DRIVERVERSION);

//    if Win32Platform = VER_PLATFORM_WIN32_NT then
//    begin
    refreshRate := GetDeviceCaps (dcHandle, VREFRESH);
    xDesktop := GetDeviceCaps (dcHandle, DESKTOPHORZRES);
    yDesktop := GetDeviceCaps (dcHandle, DESKTOPVERTRES);
//    end;

    case bitsPerPixel of
      1:   screenInfo.pixelFormat := pf1bit;
      4:   screenInfo.pixelFormat := pf4bit;
      8:   screenInfo.pixelFormat := pf8bit;
      15:  screenInfo.pixelFormat := pf15bit;
      16:  screenInfo.pixelFormat := pf16bit;
      24:  screenInfo.pixelFormat := pf24bit;
      32:  screenInfo.pixelFormat := pf32bit;
      else screenInfo.pixelFormat := pfDevice;
    end;
  end;
end;
//*********************************************************
// return size and color deep of active window as X*Y/colors
//---------------------------------------------------------
function DisplayInfoString: string;
begin
  GetDisplayInfo (screenInfo);         // get current screen device data
  with screenInfo
  do result := Format('%d*%d/%d', [x,y, bitsPerPixel]);
end;
//*********************************************************
// return size, color deep and frequence of active window
// Format:    X * Y / colors   ?? Hz
//---------------------------------------------------------
function DisplayInfoString2: string;
begin
  GetDisplayInfo (screenInfo);         // get current screen device data
  with screenInfo
  do result := Format('%d * %d / %d   %d Hz', [x,y, bitsPerPixel,refreshRate]);
end;

//*********************************************************
//  set display mode to given display mode index
//  input: index   display mode index
//                 see Windows.EnumDisplaySettings documentation
//---------------------------------------------------------
function SetDisplayMode (index: integer): longint;
var
  DevMode: TDeviceMode;
begin
  result := DISP_CHANGE_BADMODE;
  if EnumDisplaySettings (nil, index, DevMode)
  then result := ChangeDisplaySettings (DevMode, 0);
  SendMessage (HWND_BROADCAST, WM_DISPLAYCHANGE, SPI_SETNONCLIENTMETRICS, 0);
  GetDisplayInfo (screenInfo);         // get current screen device data
end;
//*********************************************************
//  set display mode to that of starting point
//---------------------------------------------------------
procedure ResetDisplayMode;
begin
  if startDisplayMode >= 0
  then SetDisplayMode (startDisplayMode);
end;
//*********************************************************
//  set display mode and display result message
//---------------------------------------------------------
procedure ChangeDisplayMode (index: integer);
var
  res: longint;
  modeTxt, resTxt: string;
  deviceMode : TDeviceMode;
begin
  res := SetDisplayMode (index);
  case res of
    DISP_CHANGE_SUCCESSFUL: resTxt := 'DISP_CHANGE_SUCCESSFUL: The settings change was successful.';
    DISP_CHANGE_RESTART   : resTxt := 'DISP_CHANGE_RESTART: The computer must be restarted in order for the graphics mode to work.';
    DISP_CHANGE_BADFLAGS  : resTxt := 'DISP_CHANGE_BADFLAGS: An invalid set of flags was passed in.';
    DISP_CHANGE_FAILED    : resTxt := 'DISP_CHANGE_FAILED: The display driver failed the specified graphics mode';
    DISP_CHANGE_BADMODE   : resTxt := 'DISP_CHANGE_BADMODE: The graphics mode is not supported';
    DISP_CHANGE_NOTUPDATED: resTxt := 'DISP_CHANGE_NOTUPDATED: Unable to write settings to the registry.';
  end;
  if EnumDisplaySettings(nil, index, deviceMode)
  then with deviceMode do modeTxt := 'mode ' + Format('%2d: %d * %d / %d   %d Hz'
     ,[index, dmPelsWidth, dmPelsHeight, dmBitsPerPel, dmDisplayFrequency]);
  ShowMessage(modeTxt + #13#10'result:'#13#10 + resTxt);
end;
//*********************************************************
//  get string list of all display modes
//  input:  colorBits   filter: number of color bits per pixel
//                      =0 use no filter
//---------------------------------------------------------
function GetDisplayModeList (colorBits: byte = 0): tStringList;

var   dm: Integer;  deviceMode: TDeviceMode;
begin
  dm := 0;
  result := tStringList.Create;
  while EnumDisplaySettings(nil, dm, deviceMode)
  do with deviceMode do begin
    if (colorBits = 0) or (colorBits = dmBitsPerPel)
    then result.Add(Format('%2d: %d * %d / %d   %d Hz'
      ,[dm, dmPelsWidth, dmPelsHeight, dmBitsPerPel, dmDisplayFrequency]));
    inc(dm);
  end;
end;
      

//=== screen handling ===

//*********************************************************
// show or hide system task bar (mostly at the screen bottom)
//---------------------------------------------------------
procedure ShowTaskBar (show: boolean);
begin
  if show
  then ShowWindow(FindWindow('Shell_TrayWnd', nil), SW_RESTORE)
  else ShowWindow(FindWindow('Shell_TrayWnd', nil), SW_HIDE);
end;


//=== drive handling ===

//*********************************************************
// return list of existing drives as string list
//---------------------------------------------------------
function GetDriveList: TStringList;
const bufSize = 256;
var   bufferPtr, helpPtr: PChar;
      strings: TStringList;
begin
  strings := TStringList.Create;
  strings.BeginUpdate;
  bufferPtr := nil;
  try
    GetMem (bufferPtr, bufSize);
    Strings.Clear;
    if GetLogicalDriveStrings (bufSize, bufferPtr) <> 0
    then begin
      helpPtr := bufferPtr;
      while helpPtr^ <> #0 do begin
        strings.Add (helpPtr);
        Inc (helpPtr, StrLen(helpPtr) + 1);
      end;
    end;
  finally
    strings.EndUpdate;
    FreeMem (bufferPtr, BufSize);
  end;
  result := strings;
end;
{*********************************************************}
{ if drive ready, return available & free disk space in KB
{ input:     drive: 0=current, 1=A, 2=B, ...}
{ output:    size, free: bytesize [Bytes]
{---------------------------------------------------------}
function DriveReady (const drive: byte; var size,free: Int64): Boolean;
var  eMode: Word;
begin
  result := false;
  size := 0;
  free := 0;
  eMode := SetErrorMode(SEM_FAILCRITICALERRORS);
  try
    size := DiskSize (drive);
    if size > 0     {drive ready ?}
    then begin
      result := true;
      free := DiskFree (drive);
    end;
  finally
    SetErrorMode(eMode);
  end;
end;
//*********************************************************
// return volume name of given drive
//---------------------------------------------------------
function GetVolumeName (drive: string): string;
var
  VolumeName    : array[0..255] of char;
  FileSystemType: array[0..255] of char;
  SerialNum, MaxFilenameLength, Flags: DWORD;
begin
  result := '';
  if (GetVolumeInformation(@drive,
                           VolumeName,256,
                           @SerialNum,
                           MaxFilenameLength,
                           Flags,
                           FileSystemType, 256))
  then result := string(VolumeName);
end;
//*********************************************************
// return list of CD/DVD drives or 'not present'
//---------------------------------------------------------
function GetCDList: shortString;
var
  AList: TStringList;
  ni, dType: integer;
  drive: shortString;
begin
  result := '';
  AList := GetDriveList;
  for ni := 0 to AList.Count-1
  do begin
    drive := Alist.Strings[ni];
    dType := GetDriveType(pchar(Alist.Strings[ni]));
    if dType = DRIVE_CDROM
    then begin
      if result <> '' then result := result + ' ';
      result := result + drive;
    end;
  end;
  if result = '' then result := 'not present';
end;


//=== file handling ===

//*********************************************************
//  set current directory to start directory
//---------------------------------------------------------
procedure RestoreStartDirectory;
begin
  SetCurrentDir (startDirectory);
end;
//*********************************************************
// get all file names of given path and extension
// example:
//   var filenameList := TStringList.Create;
//   GetFileList (filenameList, 'Data\', '*.xyz');
//---------------------------------------------------------
procedure GetFileList (var sList: tStringList; const path, ext: string);

var   sRec: TSearchRec;  hs: string;
begin
  if sList = nil
  then sList := tStringList.Create;
  try
    if FindFirst(path + ext, faAnyFile - faDirectory, sRec) = 0
    then repeat
      hs := path + sRec.Name;
      sList.Add (hs);
{$ifdef test2}
    writeln (tst, '  file=', path + sRec.Name);
{$endif}
    until FindNext(sRec) <> 0;
  finally
    SysUtils.FindClose(sRec);
    sList.Sort;
  end;
end;
//*********************************************************
// return file size of given file in bytes
//   (see also IdGlobal.FileSizeByName)
// input:      filename    name of file
// result:     integer     file size (-1 on error)
//---------------------------------------------------------
function GetFileSize (const filename: String): integer;

var    sRec: TSearchRec;
begin
  result := -1;
  if FindFirst (FileName, faAnyFile, srec) = 0 then begin
    try
      result := sRec.Size;
    finally
      SysUtils.FindClose(sRec);
    end;
  end;
end;
//*********************************************************
// return file size of given file in bytes (variant 2)
// see also IdGlobal.FileSizeByName
//---------------------------------------------------------
function FileSizeByName (const filename: string): integer;

var   pf: file of byte;
begin
  Assign (pf, filename);
  {$I-} Reset (pf); {$I+}
  ioError := ioResult;
  if ioError <> 0
  then result := -1
  else begin
    try
      result := FileSize(pf);
    finally
      CloseFile(pf);
    end;
  end;
end;
//*********************************************************
// check if file can be overwritten
// and show error message in case of an error
// note: file is closed after checking
// input:      filename    name of file
// result:     boolean     =true: file can be overwritten
//---------------------------------------------------------
function IsFileWriteable (const filename: string): boolean;

var   pf: file of byte;
begin
  Assign (pf, filename);
  {$I-} Rewrite (pf); {$I+}
  if not ShowFileError (2, filename)
  then CloseFile(pf);
  result := ioError = 0;
end;
//*********************************************************
// if drive or file is write protected for writing a file,
// a substitute filename with the temp path will be returned
// input:      filename   original file name
// result:     string     substituted file name
//---------------------------------------------------------
function SubstituteFilename (filename: string): string;
begin
  result := GetEnvironmentVariable ('temp');
  if result = ''
  then result := GetEnvironmentVariable ('tmp');
  if result = ''
  then result := 'C:\' + filename            // C:\filename
  else result := result + '\' + filename;    // C:\temp\filename
end;
//*********************************************************
// return filename extension which is the string after last dot
// see also SysUtils.ExtractFileExt
// input:      fileName        name of file
// result:     string          filename extension
// example:    ext := GetFilenameExtension ('C:\temp\test.tmp');
//                             ext = 'tmp'
//---------------------------------------------------------
function GetFilenameExtension (const filename: string): string;
var   di: Integer;
begin
  di := LastDelimiter ('.\:', FileName);
  if (di>0) and (FileName[di]='.')
  then Result := Copy (FileName, di+1, maxInt)
  else Result := '';
end;
//*********************************************************
// remove filename extension with the dot
// input:      fileName        name of file
// result:     string          filename without extension
//---------------------------------------------------------
function RemoveFilenameExt (const fileName: string): string;
var   di: Integer;
begin
  for di := Length(FileName) downto 1 do
    if FileName[di] = '\' then Break
    else if FileName[di] = '.' then
    begin
      Result := Copy(FileName, 1, di-1);
      Exit;
    end;
  Result := fileName;
end;
//*********************************************************
// copy a file using windows function e.g. to create a backup file
// see also:   IdGlobal.CopyFileTo, unit windows.CopyFile
// input:      existingFileName   name of existing file
//             newFileName        name of new file
//             failIfExists       failure handling flag
//               =FALSE if new file already exists,
//                the function overwrites the existing file
//               =TRUE and the new file already exists
//                the function fails
//---------------------------------------------------------
function CopyFile (const existingFileName: string;
                   const newFileName: string;
                   failIfExists: boolean): boolean;
begin
  result := Windows.CopyFile (pChar(@existingFileName)
                             ,pChar(@newFileName), failIfExists);
end;
//---------------------------------------------------------
procedure ShowFileProperties(const FileName: string);
var
  SEI: SHELLEXECUTEINFO;
begin
 ZeroMem (SEI,SizeOf(SEI));
 with SEI do begin
   cbSize := SizeOf(SEI);
   fMask := SEE_MASK_NOCLOSEPROCESS or SEE_MASK_INVOKEIDLIST or SEE_MASK_FLAG_NO_UI;
   Wnd := Application.Handle;
   lpVerb := 'properties';
   lpFile := PChar(FileName);
  end;
  if not ShellExecuteEx(@SEI) // then RaiseLastWin32Error; // old version
  then RaiseLastOSError;
end;


//=== program execution ===
//=== Programmstart ===

//*********************************************************
// create mutex (mutual exclusion) to prevend double program starts
// return:  = true, if program is still running
//---------------------------------------------------------
var   mutexHandle: THandle = 0;
//---------------------------------------------------------
function InitMutex (name: string): boolean;
var lastError: integer;
begin
 mutexHandle := CreateMutex(Nil, True, PChar(name));
 lastError := GetLastError;
 result := (lastError = ERROR_ALREADY_EXISTS);
end;
//---------------------------------------------------------
//  clear mutex at program end
//---------------------------------------------------------
procedure ClearMutex;
begin
  if mutexHandle <> 0 then begin
    CloseHandle(mutexHandle);
    mutexHandle := 0;
  end;
end;
//---------------------------------------------------------
// stop if program was started twice
//---------------------------------------------------------
procedure StopMultiStartedProgram;
begin
  if InitMutex (programName) then begin
    windows.Beep (440, 200);
    Halt (1);
  end;
end;

//*********************************************************
// execute DOS command
// input:      DOS command string
// example:    DOSCommand ('dir c:\*.txt');
//---------------------------------------------------------
procedure DOSCommand (command: string);
begin
//function ShellExecute(hWnd: HWND; Operation, FileName, Parameters,
//  Directory: PChar; ShowCmd: Integer): HINST; stdcall;
  ShellExecute (Application.Handle, 'open'
    ,PChar(SysUtils.GetEnvironmentVariable('COMSPEC'))
//    ,PChar('/C ' + command), nil, SW_Show);
    ,PChar(command), nil, SW_Show);
end;
//*********************************************************
// execute Windows application
// input:    name: string     application name
//           param: string    application start parameters
// result:   boolean          =true application startet
// errorCode
//---------------------------------------------------------
function RunApplication (const name, param: string): boolean;

var   errorCode: cardinal;
begin
  errorCode := ShellExecute (Application.Handle, 'open'
     ,PChar(name), nil, PChar(param), SW_SHOWNA);   //sw_ShowNormal);
  result := errorCode <> 0;
  if not result then
    ShowError ('SysUtil2: RunApplication'
              ,'can not start application ''' + name + '''' + CRLF, MB_OK);
end;
//*********************************************************
// execute system control panel applet
// input:      appletFileName
//               access.cpl input
//               appwiz.cpl software list
//               desk.cpl   screen display
//               intl.cpl   language
//               joy.cpl    joystick
//               main.cpl   mouse
//               mmsys.cpl  multimedia / sound
//               modem.cpl  modems
//               sscpl.cpl  sound system
//               sysdm.cpl  system
//               timedate.cpl date/time
// example:    RunControlPanelApplet('timedate.cpl');
//---------------------------------------------------------
procedure RunControlPanelApplet (appletFileName: string);
begin
  ShellExecute(Application.handle, 'open', PChar('rundll32.exe'),
    PChar('shell32.dll,Control_RunDLL '+ appletFileName), nil, SW_SHOWNORMAL);
end;
//*********************************************************
// execute internet link
// input:      link address
// examples:   RunLink ('http://www.google.de');
//             RunLink ('mailto:name@xxx.com?subject=title');
//---------------------------------------------------------
procedure RunLink (link: string);
begin
  ShellExecute(0, Nil, PAnsiChar(link), Nil, Nil, SW_NORMAL);
end;


//=== display message and errors ===
//=== Anzeige von Meldungen und Fehlern ===

//*********************************************************
// raise error message, that object pointer is nil
// (correspond to RuntimeError 222)
//---------------------------------------------------------
// Ausgeben der Fehlermeldung, da Objektzeiger = 0 ist
// (entspricht RuntimeError 222)
//---------------------------------------------------------
procedure RaiseObjectNilError (name: string);
begin
  Raise EInvalidPointer.Create (#13#10'Object pointer = nil when calling ' +
    name + #13#10);
end;
//*********************************************************
// raise error message, that object isn't initialized
// (correspond to RuntimeError 210)
//---------------------------------------------------------
// Ausgeben der Fehlermeldung, da Objekt nicht initialisiert ist
// (entspricht RuntimeError 210)
//---------------------------------------------------------
procedure RaiseObjectInitError (name: string);
begin
  Raise EInvalidPointer.Create (#13#10'Object of type "' +
    name + '" is not initialized'#13#10);
end;
//*********************************************************
// raise error message, that a heap memory error occured
// (correspond to RuntimeError 217)
//---------------------------------------------------------
// Ausgeben der Fehlermeldung, da ein Heap-Speicherfehler aufgetreten ist
// (entspricht RuntimeError 217)
//---------------------------------------------------------
procedure RaiseMemoryError (name: string);
begin
  Raise EInvalidPointer.Create (#13#10'no heap memory at "' +
    name + '"'#13#10);
end;
//*********************************************************
// raise error message, that a procedure pointer
// (correspond to RuntimeError 221)
//---------------------------------------------------------
// Ausgeben der Fehlermeldung, da Procedurfehler aufgetreten ist
// (entspricht RuntimeError 221)
//---------------------------------------------------------
procedure RaiseProcedureError (name: string);
begin
  Raise EInvalidPointer.Create (#13#10'no heap memory at "' +
    name + '"'#13#10);
end;

//*********************************************************
// return last window error message
// result:     string     last windows error message
//---------------------------------------------------------
// Text des letzten Windows-Fehlers zurckgeben
// Ergebis:    string     Fehlermeldung in Textform
//---------------------------------------------------------
function Win32ErrorString(code: Integer): String;
var   lpMsgBuf: PChar;
begin
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER or
                FORMAT_MESSAGE_FROM_SYSTEM or
                FORMAT_MESSAGE_IGNORE_INSERTS,
                nil, code, 0, @lpMsgBuf, 0, nil);
  result := String(lpMsgBuf);
  LocalFree(Cardinal(lpMsgBuf));
end;

//*********************************************************
// show message in modal window
//---------------------------------------------------------
// Meldung in modalem Fenster anzeigen
//---------------------------------------------------------
procedure ShowMessageBox (const caption, infoText: string);
begin
  msgBoxIsOpen := true;
  Application.MessageBox (PChar(infoText), PChar(caption));
  msgBoxIsOpen := false;
end;
//*********************************************************
// show hint message in modal window
//---------------------------------------------------------
// Hinweismeldung in Fenster anzeigen
//---------------------------------------------------------
procedure ShowHint (const caption, infoText: string);
begin
  msgBoxIsOpen := true;
  Application.MessageBox (PChar(infoText), PChar(caption)
                         , mb_Ok or MB_ICONINFORMATION);
  MarkTime;                  // restart time measuring
  msgBoxIsOpen := false;
end;
//*********************************************************
// show warning message in modal window
// and copy message text to clipboard
//---------------------------------------------------------
// Warnungsmeldung in Fenster anzeigen
// und Fehlertext ins Clipboard kopieren
//---------------------------------------------------------
procedure ShowWarning (const caption, infoText: string);
begin
  msgBoxIsOpen := true;
  if Application.MessageBox (PChar(infoText), PChar(caption)
                            ,mb_OkCancel or MB_ICONWARNING) = IDCANCEL
  then Halt (errUserBreak);
  Clipboard.SetTextBuf(PAnsiChar(infoText));   // copy text to clipboard
  MarkTime;                // restart time measuring
  msgBoxIsOpen := false;
end;
//*********************************************************
// Show error message in modal window
// and copy message text to clipboard.
// On debug mode error message will be written to debug file.
//---------------------------------------------------------
// Fehlermeldung in modalem Fenster anzeigen
// und Fehlertext ins Clipboard kopieren
// Im Debugmodus wird die Fehlermeldung auch im Debugfile eingetragen
// Versorgung:  caption   berschrift ('Modulname.Prozedurname')
//              infoText  Beschreibung der Fehlerursache
//              buttons   optionale Angabe der Buttons
//                        MB_OK, MB_OKCANCEL, ...
//                        $102 = MB_ABORTRETRYIGNORE + MB_DEFBUTTON2
//                        MB_ICONERROR = mit error Icon + OK button
//                        MB_ICONERROR = mit error Icon + OK button
// Rckgabe:    Buttonwert von Application.MessageBox
//                3 = ID_ABORT
//                4 = ID_RETRY
//                5 = ID_IGNORE
// Beispiele:
// ShowError ('unit.procedure', 'infotext');
// if ShowError ('unit.procedure', 'info', MB_OKCANCEL) <> ID_RETRY then close;
//---------------------------------------------------------
function ShowError (const caption, infoText: string;
                    const buttons: word = $102): integer;
//var  s1: string;
begin
{$ifopt D+}
  writeln (tst,'>>> ShowError: ', caption, #13#10'    ', infoText);
  flush (tst);
{$endif}
  result := 0;
  msgBoxIsOpen := true;
  try
//    s1 := WrapText (infoText, #13#10, ['.',','] , 30);
//    result := MessageDlg (PChar(infoText), mtError, [mbOk], 0);
    result := Application.MessageBox (PChar(infoText)
              ,PChar(caption), buttons or MB_ICONERROR);
    case result of                       // set debug stop here !!!
      //IDCANCEL: Application.Terminate;
      IDCANCEL, IDABORT: begin
        Clipboard.SetTextBuf(PAnsiChar(infoText));
        Halt (errUserBreak);
      end;
      // IDNO, IDRETRY, IDIGNORE: ;
    end;
  except
    Halt (errFatalError);
  end;
  msgBoxIsOpen := false;
  MarkTime;
end;
//*********************************************************
// Letzten Windows-Fehler im Klartext in modalem Fenster
// anzeigen (fr Eingabe und Ergebnis siehe ShowError)
// Wenn LastError = 0 wird keine Meldung angezeigt !
//---------------------------------------------------------
// show last window error message in modal window
// (for input and result see ShowError)
// if LastError = 0, no message will be shown !
//---------------------------------------------------------
function ShowLastError (const caption, infoText: string;
                        const buttons: word = $102): integer;

var  s1,s2: string;   err: Integer;
begin
  msgBoxIsOpen := true;
  err := GetLastError;
  if err = 0
  then s1 := infotext
  else s1 := infotext + #13#10 + Format ('  error %d: %s', [err, Win32ErrorString(err)]);
{$ifopt D+}
  writeln (tst,'>>> ShowLastError: ', caption, #13#10'  ', s1);
  flush (tst);
{$endif}
  result := IDIGNORE;
  if err = 0 then exit;                //===>
  try
    s2 := ' Error in ' + caption + #0;
    result := Application.MessageBox (PChar(s1), PChar(s2), buttons);
    case result of                     // set debug stop here !!!
      //IDCANCEL: Application.Terminate;
      IDCANCEL, IDABORT: Halt (errUserBreak);
      // IDNO, IDRETRY, IDIGNORE: ;
    end;
  except
    Halt (errFatalError);
  end;
  msgBoxIsOpen := false;
  MarkTime;
end;
//*********************************************************
// show fatal error message in modal window
//---------------------------------------------------------
// Fehlermeldungen ausgeben und Programm beenden.
// Versorgung:  exitCode  Beendingungscode 0..255
//              caption   berschrift ('Modulname.Prozedurname')
//              infoText  Beschreibung der Fehlerursache
//---------------------------------------------------------
procedure ShowFatalError (const exitCode: byte;
                          const caption, infoText: string);

var  s2: string[255];
begin
  s2 := 'error in ' + caption + #13#10#13#10 + infoText + #13#10#0;
{$ifopt D+}
  writeln (tst, '>>> ShowFatalError: ', exitcode, ', ', s2);
  flush (tst);
{$endif}
  try
    MessageDlg (s2, mtError, [mbAbort], 0);
    Halt (exitcode);                     // set debug stop here !!!

  except
    Halt (errFatalError);
  end;
end;
//*********************************************************
const fileErrorMsg: array [100..107] of string
      =('Fehler beim Lesen vom Datentrger',          // 100
        'Fehler beim Schreiben auf Datentrger',      // 101
        'Dateivariable ist keiner Datei zugewiesen',  // 102
        'Datei nicht geffnet',                       // 103
        'Datei nicht fr Eingabe geffnet',           // 104
        'Datei nicht fr Ausgabe geffnet',           // 105
        'Ungltiges Eingabeformat',                   // 106
        'Datei ist bereits geffnet');                // 107

//*********************************************************
// show file error message in modal window if error <> hide
//---------------------------------------------------------
// Anzeige des Dateifehlers, wenn der Fehler ungleich hide ist.
// Delphi-Hilfe: EInOutError, IOResult-->E/A-Fehler
// Beispiel:     if ShowFileError (2, 'test.txt') then exit; 
//---------------------------------------------------------
function ShowFileError (hide: byte; filename: string): boolean;
var   ft: string;
begin
  ShowFileError := false;
  ioError := ioResult;
  if (ioError = 0) or (ioError = hide) then exit;          //===>
  ft := 'error ' + IntToStr (ioError) + ' at file "' + filename + '"'#13#10;
  if (ioError < 100) or (ioError > 107)
  then ft := ft + SysErrorMessage (ioError)
  else ft := ft + fileErrorMsg[ioError];

  if MessageDlg (PChar(ft), mtError, [mbOk,mbCancel], 0) = IDCANCEL
  then Halt (errUserBreak);

{$ifopt D+}
  // Achtung: Debug-Ausgabe wird nach MessageDlg aufgerufen,
  // da ansonsten Fehlermeldung beim Schreiben auf Debug-Datei berschrieben
  // wird, wenn Debug-Datei nicht geffnet werden kann.
  if filename <> debugFilename then begin
    writeln (tst, '>>> ShowFileError: ', ft);
    flush (tst);
  end;
{$endif}

  ShowFileError := true;
end;

//*********************************************************
// open test file for debug output
//---------------------------------------------------------
// ffnet Testdatei fr Debug-Ausgaben
//---------------------------------------------------------
{$ifopt D+}
procedure OpenTestFile;
var filename: string;
begin
  filename := debugFilename;
  Assign (tst, filename);
  {$I-} ReWrite (tst); {$I+}         // Datei erstellen
  if ShowFileError (0, debugFilename)
  then begin
    // try to create debug file in temp directory
    filename := SubstituteFilename(filename);
    MessageDlg (PChar('using testfile ' + filename)
               , mtInformation, [mbOk], 0);
    Assign (tst, filename);
    {$I-}   ReWrite (tst);   {$I+}
  end;
  if ShowFileError (0, filename) then Halt (ioError);   // Start abbrechen

  writeln (tst, debugFilename, ': Delphi Debug File'#13#10
              , '  application: ', Application.ExeName, #13#10
              , '  time: ', DateTimeToStr(Now), #13#10);
  Flush (tst);
end;
{$endif}



//=========================================================
//  Stack-Routinen
//=========================================================

//*********************************************************
// Stapelbearbeitung fr Integer-Werte
//---------------------------------------------------------
const integerStackStep = 64;
var   integerStack: array of integer;
      integerStackLen: word = 128;
      integerMaxLen: word = 1024;
      integerStackIndex: word = 0;
//---------------------------------------------------------
procedure InitIntegerStack;
begin
  SetLength (IntegerStack, IntegerStackLen);
  IntegerStackIndex := 0;
end;
//---------------------------------------------------------
// Kopie eines Long-Werte auf den Stapel kopieren
//---------------------------------------------------------
function PushInteger (const value: Integer): Integer;
begin
  IntegerStack[IntegerStackIndex] := value;
  if integerStackIndex >= integerMaxLen
  then begin
    {$ifopt D+}
      ShowWarning ('SysUtil2: PushInteger', 'integer stack is full: check call of PopInteger');
    {$endIf}
  end else begin
    inc (IntegerStackIndex);
    if IntegerStackIndex >= IntegerStackLen
    then begin
      inc (IntegerStackLen, IntegerStackStep);
      SetLength (IntegerStack, IntegerStackLen);
    end;
  end;
  result := value;
end;
//---------------------------------------------------------
// Kopie eines Long-Wertes vom Stapel nehmen
//---------------------------------------------------------
function PopInteger: integer;
begin
  if IntegerStackIndex > 0
  then dec (IntegerStackIndex);
  result := IntegerStack[IntegerStackIndex];
end;

//*********************************************************
// Stapelbearbeitung fr Boolean-Werte
//---------------------------------------------------------
const booleanStackStep = 64;
var   booleanStack: array of boolean;
      booleanStackLen: word = 128;
      booleanMaxLen: word = 1024;
      booleanStackIndex: word = 0;
//---------------------------------------------------------
procedure InitBooleanStack;
begin
  SetLength (booleanStack, booleanStackLen);
  booleanStackIndex := 0;
end;
//---------------------------------------------------------
// Kopie eines Boolean-Werte auf den Stapel kopieren
//---------------------------------------------------------
function PushBoolean (const value: boolean): boolean;
begin
  booleanStack[booleanStackIndex] := value;
  if booleanStackIndex >= booleanMaxLen
  then begin
    {$ifopt D+}
      ShowWarning ('SysUtil2: PushBoolean', 'boolean stack is full: check call of PopBoolean');
    {$endIf}
  end else begin
    inc (booleanStackIndex);
    if booleanStackIndex >= booleanStackLen
    then begin
      inc (booleanStackLen, booleanStackStep);
      SetLength (booleanStack, booleanStackLen);
    end;
  end;
  result := value;
end;
//---------------------------------------------------------
// Kopie eines Boolean-Wertes vom Stapel nehmen
//---------------------------------------------------------
function PopBoolean: boolean;
begin
  if booleanStackIndex > 0
  then dec (booleanStackIndex);
  result := booleanStack[booleanStackIndex];
end;

//*********************************************************
// Stapelbearbeitung fr Float-Werte
//---------------------------------------------------------
const floatStackStep = 64;
var   floatStack: array of float;
      floatStackLen: word = 128;
      floatMaxLen: word = 1024;
      floatStackIndex: word = 0;
//---------------------------------------------------------
procedure InitFloatStack;
begin
  SetLength (floatStack, floatStackLen);
  floatStackIndex := 0;
end;
//---------------------------------------------------------
// Kopie eines Float-Werte auf den Stapel kopieren
//---------------------------------------------------------
function PushFloat (const value: float): float;
begin
  floatStack[floatStackIndex] := value;
  if floatStackIndex >= floatMaxLen
  then begin
    {$ifopt D+}
      ShowWarning ('SysUtil2: PushFloat'
                  ,'float stack is full: check call of PopFloat');
    {$endIf}
  end else begin
    inc (floatStackIndex);
    if floatStackIndex >= floatStackLen
    then begin
      inc (floatStackLen, floatStackStep);
      SetLength (floatStack, floatStackLen);
    end;
  end;
  result := value;
end;
//---------------------------------------------------------
// Kopie eines Float-Wertes vom Stapel nehmen
//---------------------------------------------------------
function PopFloat: float;
begin
  if floatStackIndex > 0
  then dec (floatStackIndex);
  result := floatStack[booleanStackIndex];
end;

//=== integer arithmetic ===

{$R-,Q-}
//*********************************************************
// calculate square value of given inter value
// error message of under- or overflow will be suppressed
//---------------------------------------------------------
// Quadratwert des angegebenen Ganzzahlwertes berechnen.
// Fehlermeldung bei Unter- oder berlauf wird unterdrckt.
//---------------------------------------------------------
function SqrInt (const value: Integer): Integer;
begin
  {$ifdef VER150}  // Delphi 7.0
    if value < minShort                // -32768
    then result := integer(minLong)    // -2147483648
    else if value > maxShort           // 32767
    then result := maxLong             // 2147483647
    else result := value * value;
  {$else}
    if value < -32768                  // minShort
    then result := -2147483648         // minLong
    else if value > 32767              // maxShort
    then result := 2147483647          // maxLong
    else result := value * value;
  {$endif}
end;
//*********************************************************
// calculate square root of given integer value
//---------------------------------------------------------
// Quadratwurzel des angegebenen Ganzzahlwertes berechnen.
//---------------------------------------------------------
function SqrtInt (const value: Integer): Integer;
begin
  result := round (sqrt(value));
end;
//*********************************************************
// computes a*b/c with an intermediate 64 bits precision.
//---------------------------------------------------------
// Berechnet a*b/c mit 64 Bit Genauigkeit
//---------------------------------------------------------
function MulDiv (const a, b, c: Integer): Integer;
begin
  result := Int64(a) * b div c;
end;
//---------------------------------------------------------
// note: integer overflow if a*b > maxlong
//---------------------------------------------------------
(*
function MulDiv (const a, b, c: Integer): Integer;
asm
  imul   edx
  idiv   ecx
end;
*)


//===================================================================
// class TObjectStack: replacement of class Contnrs.TObjectStack
// is faster because of compact coding of the class.
//---------------------------------------------------------
// Klasse TObjectStack: Ersatz der Klasse Contnrs.TObjectStack
// ist etwas schneller wegen der kompakten Kodierung der Klasse.
//===================================================================

//*********************************************************
// constructor creates an instance of TObjectStack
//---------------------------------------------------------
// Der Konstruktor erzeugt eine Instanz von TObjectStack.
//---------------------------------------------------------
constructor TObjectStack.Create;
begin
{$ifdef stackTest}
  writeln (tst, HexString(self), ' TObjectStack.Create: ');
{$endif}
  inherited Create;
  FList := TList.Create;
end;
//*********************************************************
// destructor erases own instance of TObjectStack
//---------------------------------------------------------
// Der Destruktor lscht eigene Instanz von TObjectStack.
//---------------------------------------------------------
destructor TObjectStack.Destroy;
begin
{$ifdef stackTest}
  writeln (tst, HexString(self), ' TObjectStack.Destroy: ');
{$endif}
  FList.Free;
  inherited Destroy;
end;
//*********************************************************
// clear all list objects
//---------------------------------------------------------
// Lscht alle Objekte des Stapels
//---------------------------------------------------------
procedure TObjectStack.Clear;
begin
  FList.Clear;
{$ifdef stackTest}
  writeln (tst, HexString(self), ' TObjectStack.Clear: ',FList.Count);
{$endif}
end;
//*********************************************************
// copy object reference onto stack
// input:      aObject;   Object
// result:     tObject;   reference to added object
//---------------------------------------------------------
// Kopie einer Objektreferenz auf dem Stapel kopieren
// Versorgung: aObject;   Objekt
// Ergebnis:   tObject;   Referenz auf hinzugefgtes Objekt
//---------------------------------------------------------
function TObjectStack.Push (AObject: tObject): tObject;
begin
{$ifopt D+}
  if AObject = nil then RaiseObjectNilError  ('TObjectStack.Push');
{$endif}
  FList.Add(AObject);        // copy object reference onto stack
  result := AObject;
{$ifdef stackTest}
  writeln (tst, HexString(self), ' TObjectStack.Push: ',FList.Count);
{$endif}
end;
//*********************************************************
// remove last object reference from stack and return it
// result:     tObject;   removed object
//---------------------------------------------------------
// Entfernt oberste Objektreferenz vom Stapel und gibt diese zurck
// Ergebnis:   tObject;   entferntes Objekt
//---------------------------------------------------------
function TObjectStack.Pop: tObject;
begin
{$ifdef stackTest}
  writeln (tst, HexString(self), ' TObjectStack.Pop: ',FList.Count-1);
  flush (tst);
{$endif}

  if FList.Count = 0
  then begin
    result := nil;   exit;           //===>
  end;

  //-- result := PopItem; --
  result := FList[FList.Count-1];    // Objektreferenz
  FList.Delete(FList.Count-1);       // Objekt lschen
end;
//*********************************************************
// return reference to last stack object
// result:     tObject     object reference
//---------------------------------------------------------
// Gibt Referenz auf das oberste Objekt des Stapels zurck
// Ergebnis:   tObject     Objektreferenz
//---------------------------------------------------------
function TObjectStack.Peek: tObject;
begin
{$ifdef stackTest}
  writeln (tst, HexString(self), ' TObjectStack.Peek: ',FList.Count);
{$endif}
  result := FList[FList.Count-1];   // result := PeekItem;
end;
//*********************************************************
// return true, if stack count is greater or equal aCount
// result:     boolean    = true, if count >= aCount
//---------------------------------------------------------
// gibt True zurck, wenn Anzahl der Listenelemente >= aCount ist.
// Ergebnis:   boolean    =true, wenn Count >= aCount
//---------------------------------------------------------
function TObjectStack.AtLeast (aCount: integer): boolean;
begin
  Result := FList.Count >= aCount;
end;
//*********************************************************
// return number of stack objects
// result:     integer;      number of stack objects
//---------------------------------------------------------
// Liefert die Anzahl der Stackelemente
// Ergebnis:   integer;      Objektanzahl des Stapels
//---------------------------------------------------------
function TObjectStack.Count: integer;
begin
  result := FList.Count;
end;
//*********************************************************
// put given element on the stack
// input:      aItem;
//---------------------------------------------------------
// Legt ein Element auf den Stapel
// Versorgung: aItem;   Element
//---------------------------------------------------------
procedure TObjectStack.PushItem (aItem: Pointer);
begin
  FList.Add(aItem);
end;
//*********************************************************
// return pointer to the last added stack item
// result:   pointer
//---------------------------------------------------------
// Gibt einen Zeiger auf das letzte Listenelement zurck
// Ergebnis: pointer
//---------------------------------------------------------
function TObjectStack.PeekItem: pointer;
begin
  result := FList[FList.Count-1];
end;
//*********************************************************
// delete last added item and return it's pointer
// result:   pointer;     pointer to deleted stack item
//---------------------------------------------------------
// Entfernt letztes Listenelement und gibt Zeiger auf diesen zurck.
// Ergebnis: pointer;     Zeiger auf entferntes Listenelement
//---------------------------------------------------------
function TObjectStack.PopItem: Pointer;
begin
  result := FList[FList.Count-1];
  FList.Delete(FList.Count-1);
end;



{$ifopt D+}
//*********************************************************
// only for testing
//---------------------------------------------------------
// nur fr den Test
//---------------------------------------------------------
procedure Test1;             // test BinString, HexString
var  vb: byte;   vw: word;   vl: longword;  vi: int64;
begin
  vb := $AB;
  vw := $ABCD;
  vl := $ABCD1234;
  vi := $ABCD1234567890EF;
(*                                // standard convertion
  writeln (tst, vb:11, '  $'+IntToHex(vb,8):11);
  writeln (tst, vw:11, '  $'+IntToHex(vw,8):11);
  writeln (tst, vl:11, '  $'+IntToHex(vl,8):11);
  writeln (tst, vi:11, '  $'+IntToHex(vi,8):11);
*)
  writeln (tst, vb:11, '  $'+HexString(vb):11,'  ',BinString(vb));
  writeln (tst, vw:11, '  $'+HexString(vw):11,'  ',BinString(vw));
  writeln (tst, vl:11, '  $'+HexString(vl):11,'  ',BinString(vl));
  writeln (tst,        '  $'+HexString(vi):22,'  ',BinString(vi));
  writeln (tst);
end;
//---------------------------------------------------------
procedure Test2;
var   lv: longint;   bv: byte;   src, dst: array[0..2] of longint;
begin
  lv := 200000;
  lv := Clamp (lv, -100, 123456);
  lv := ClampByte (lv);
  bv := ClampByte (-lv);
  // SAR
  lv := bv + SAR_2(-16);
  bv := byte (lv + SAR_15(-123456));
  // memory test
  FillLongWord (src, 3, 123456+bv);
  MoveLongWord (src, dst, 3);
  for lv := 0 to 2 do src[lv] := 100000+lv;
  SwapMem (src, dst, 3*sizeof(longint));
end;
//---------------------------------------------------------
procedure Test3;        // test string handling
var   s1: string;  //br: boolean;
begin
  s1 := TrimAll (' 12 34 ');             //='1234'
  s1 := TrimAll ('.1.2.3.4.', '.');      //='1234'
  s1 := TrimAll ('_+_12_+_34', '_+_');   //='1234'
  s1 := TrimAll ('****12***34**', '*');  //='1234'
(*
  br := IsInteger ('0');
  br := IsInteger ('1234567890');
  br := IsInteger ('12345678901');
  br := IsInteger ('-1234');
  br := IsInteger ('+12345678912345670');    // false
  br := IsInteger ('+999999999');            // false
  br := IsInteger ('+-1234');                // false

  br := IsInt64 ('+1234');
  br := IsInt64 ('+12345678901234');
  br := IsInt64 ('+12345678901234567890');   // false

  br := IsFloat ('1');
  br := IsFloat ('1.23456');
  br := IsFloat ('1,23456');
  br := IsFloat ('-12.34E-56');
  br := IsFloat ('1,23E+45');
  br := IsFloat ('1.234E+56');
  br := IsFloat ('-12,34E+34567');     // false
  br := IsFloat ('-1,2E-34x');         // false
  br := IsFloat ('--1,2E-34');         // false
  br := IsFloat ('-+1,2E+34');         // false
  br := IsFloat ('-1,2E+3,4');         // false
  br := IsFloat ('-123.45,6');         // false
*)
end;
//---------------------------------------------------------
{$R+,Q+}
procedure Test4;      // test integer arithmetic
begin
  writeln (tst, MulDiv (1234567890, 222, 222));       // nur mit int64 richtig
  writeln (tst, MulDiv (maxLongInt, 65000, 65000));
  writeln (tst, MulDiv (222222, -maxLongInt, -maxLongInt));
  writeln (tst, MulDiv (maxLong, 4, 3));
  writeln (tst, SqrtInt (maxLong), sqrt(maxLong):20:2);
  writeln (tst, SqrtInt (SqrInt (4444)), Sqrt (SqrInt (4444)):20:2);
  writeln (tst, 'maxLong=', IntDotString(maxLong));
  writeln (tst, 'SqrInt(111111)=',SqrInt (111111):12,' ',IntDotString(SqrInt (111111)));
//  writeln (tst, 'Sqr   (111111)=',Sqr    (111111):12,' ',IntDotString(Sqr    (111111)));
  writeln (tst);
end;
//---------------------------------------------------------
procedure Test5;
begin
  writeln (tst, 'min=',Min (-22,44), '  max=',Max (-22,44));
  writeln (tst, 'min=',Min (-222222,0,444444), '  max=', Max (-222222,0,444444));
  writeln (tst);
end;
//---------------------------------------------------------
procedure FileTest;
var   filenameList: tStringList;   filename, fullname: string;
begin
  writeln (tst,  'startDirectory: ', startDirectory);
  filenameList := tStringList.Create;
  GetFileList (filenameList, '', '*.exe');
  filename := filenameList[0];
  fullname := startDirectory + '\' + filename;
  writeln (tst,  'filename = ', filename);
  writeln (tst,  'filesize = ',  IntDotString (GetFileSize (filename)));
  writeln (tst,  'filesize = ',  IntDotString (FileSizeByName (filename)));
  writeln (tst,  'substitute = ',  SubstituteFilename (filename));
  writeln (tst,  'file ext = ', GetFilenameExtension (fullname));
  writeln (tst,  'without ext = ', RemoveFilenameExt (fullname));
  writeln (tst,  'writeable = ', IsFileWriteable (fullname));
  ShowFileProperties (filename);
end;
//*********************************************************
type  tMixed = class(TObject)
        iv: integer;
        sv: string;
        fv: float;
      end;
//---------------------------------------------------------
procedure StackTest;

var   ni: integer;   testStack: TObjectStack;   testobject: tMixed;
begin
  writeln (tst,  'StackTest: ');
  testStack := TObjectStack.Create;    // Stapel erzeugen
  // erzeugen
  for ni := 1 to 3 do begin
    testobject := tMixed.Create;       // neues Objekt erzeugen
    with testobject do begin
      iv := ni;
      sv := IntString(ni);
      fv := ni + ni/100;
    end;
    testStack.Push (testobject);       // Objekt auf Stapel legen
  end;
  testobject.iv := 999;
  // lesen
  writeln (tst, 'TestStack.count=',testStack.Count);
  for ni := 1 to testStack.Count
  do begin
    testobject := tMixed(testStack.Pop);  // Objekt vom Stapel nehmen
    with testobject do
    writeln (tst, iv:2, '  s=', sv, '  f=',fv:0:1);
    testObject.Free;                   // Objekt lschen
  end;
  // lschen
  testStack.Free;                      // Stapel lschen
end;
{$endif}

//*********************************************************
initialization

  {$ifopt D+}
    OpenTestFile;
  {$endif}

  programName := GetProgramName;
  computerName := GetComputerName;
  startDirectory := GetCurrentDir;
  MemStat := MemoryStatus;
  InitDisplayInfo;
  InitIntegerStack;
  InitBooleanStack;
  InitFloatStack;
  InitSound;

  MarkTime;
  InitTicker;

// use this test calls to check system functions
//  Test1;
//  Test2;
//  Test3;
//  Test4;
//  Test5;
//  FileTest;
//  DOSCommand ('dir c:\*.txt');    // for testing
//  StackTest;

//---------------------------------------------------------
finalization
  {$I-}
  {$ifopt D+}
    Close (tst);             {close debug file}
    ioError := ioResult;
  {$endif}
  ClearMutex;
  // ResetDisplayMode;
end.

